"""
Neuraxle's Base Hyperparameter Repository Classes
=================================================
Data objects and related repositories used by AutoML.

Classes are splitted like this for the AutoML:
- Projects
- Clients
- Rounds (runs)
- Trials
- TrialSplits
- MetricResults

..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
import copy
import datetime
import json
import logging
import os
import typing
from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass, field, fields
from json.encoder import JSONEncoder
from typing import (Any, Dict, Generic, List, Optional, Sequence, Tuple, Type,
                    TypeVar, Union)

import neuraxle.metaopt.data.aggregates as agg
import numpy as np
from neuraxle.base import CX, BaseService, ContextLock, TrialStatus
from neuraxle.hyperparams.space import HyperparameterSamples, RecursiveDict
from neuraxle.logging.logging import (LOGGING_DATETIME_STR_FORMAT,
                                      NeuraxleLogger)
from neuraxle.logging.warnings import RaiseDeprecatedClass
from neuraxle.metaopt.observable import _ObservableRepo

SubDataclassT = TypeVar('SubDataclassT', bound=Optional['BaseDataclass'])
ScopedLocationAttrInt = int
ScopedLocationAttrStr = str
ScopedLocationAttr = Union[str, int]
Slice = Sequence

DEFAULT_PROJECT: ScopedLocationAttrStr = "default_project"
DEFAULT_CLIENT: ScopedLocationAttrStr = "default_client"
DEFAULT_ROUND: ScopedLocationAttrInt = 0
DEFAULT_TRIAL: ScopedLocationAttrInt = 0
DEFAULT_TRIAL_SPLIT: ScopedLocationAttrInt = 0
DEFAULT_METRIC_NAME: ScopedLocationAttrStr = "main"


@dataclass(order=True)
class ScopedLocation(BaseService):
    """
    A location in the metadata tree.
    """
    project_name: ScopedLocationAttrStr = None
    client_name: ScopedLocationAttrStr = None
    round_number: ScopedLocationAttrInt = None
    trial_number: ScopedLocationAttrInt = None
    split_number: ScopedLocationAttrInt = None
    metric_name: ScopedLocationAttrStr = None

    def __post_init__(self):
        BaseService.__init__(self)

    # get the field value from BaseMetadata subclass:
    def __getitem__(
        self, key: Union[Type['BaseDataclass'], Slice[Type['BaseDataclass']]]
    ) -> Union[ScopedLocationAttr, 'ScopedLocation']:
        """
        Get sublocation attr from the provided :class:`BaseMetadata` type,
        or otherwise get a slice of the same type, sliced from a :class:`BaseMetadata` type range of attributes to keep.
        If the key is a slice, the end loc is kept as inclusive.
        """
        if key is None:
            return ScopedLocation()

        if isinstance(key, slice):
            if key.stop is None or key.start is not None:
                raise ValueError("Slice stop must be specified and slice start must be None.")

            idx: int = key.stop
            if not isinstance(idx, int):
                if idx == RootDataclass:
                    return ScopedLocation()
                idx: int = list(dataclass_2_id_attr.keys()).index(key.stop) + 1

            return ScopedLocation(
                *self.as_list()[:idx]
            )

        if key == RootDataclass:
            return None

        if isinstance(key, int):
            if key < 0:
                if len(self) == 0:
                    return None
                else:
                    return self.as_list()[key]
            key = list(dataclass_2_id_attr.keys())[key]

        if key in dataclass_2_id_attr.keys():
            return getattr(self, dataclass_2_id_attr[key])

        raise ValueError(f"Invalid key type {key.__class__.__name__} for key {key}.")

    def copy(self) -> 'ScopedLocation':
        """
        Returns a copy of the :class:`ScopedLocation`.
        """
        return ScopedLocation(*self.as_list())

    def with_dc(self, dc: 'BaseDataclass') -> 'ScopedLocation':
        """
        Returns a new :class:`ScopedLocation` with the provided :class:`BaseDataclass` (dc) type's id added.
        """
        if isinstance(dc, RootDataclass):
            return ScopedLocation()
        self_copy = self.copy()
        self_copy[dc.__class__] = dc.get_id()
        return self_copy

    def with_id(self, _id: ScopedLocationAttr) -> 'ScopedLocation':
        """
        Returns a longer :class:`ScopedLocation` with the provided id added at the end.
        """
        return ScopedLocation(*(self.as_list() + [_id]))

    def at_dc(self, dc: 'BaseDataclass') -> 'ScopedLocation':
        """
        Returns a trimmed :class:`ScopedLocation` with the provided :class:`BaseDataclass` (dc) type's id as the ScopedLocation's deepest attribute.
        """
        if isinstance(dc, RootDataclass):
            return ScopedLocation()
        return self[:dc.__class__]

    @staticmethod
    def default_full() -> 'ScopedLocation':
        """
        Returns a :class:`ScopedLocation` with all attributes
        set to the default non-null value instead of None.
        """
        return ScopedLocation(
            DEFAULT_PROJECT, DEFAULT_CLIENT, DEFAULT_ROUND, DEFAULT_TRIAL, DEFAULT_TRIAL_SPLIT, DEFAULT_METRIC_NAME)

    @staticmethod
    def default(
        round_number: Optional[ScopedLocationAttrInt] = None,
        trial_number: Optional[ScopedLocationAttrInt] = None,
        split_number: Optional[ScopedLocationAttrInt] = None,
        metric_name: Optional[ScopedLocationAttrStr] = None,
    ) -> 'ScopedLocation':
        """
        Returns the default :class:`ScopedLocation`. That is:

        .. code-block:: python
            ScopedLocation("default_project", "default_client", *kargs, **kwargs).

        """
        if not ((round_number is not None) >= (trial_number is not None) >= (split_number is not None) >= (metric_name is not None)):
            raise ValueError(
                "round_number, trial_number, split_number, and metric_name "
                "must be specified in order, one after the other.")

        _args = []
        if round_number is not None:
            _args.append(round_number)
            if trial_number is not None:
                _args.append(trial_number)
                if split_number is not None:
                    _args.append(split_number)
                    if metric_name is not None:
                        _args.append(metric_name)

        return ScopedLocation(DEFAULT_PROJECT, DEFAULT_CLIENT, *_args)

    def __setitem__(self, key: Type['BaseDataclass'], value: ScopedLocationAttr):
        """
        Set sublocation attr from the provided :class:`BaseMetadata` type.
        """
        # Throw value error if the key's type is not yet to be defined:
        curr_attr_to_set_idx: int = len(self)
        key_idx: int = list(dataclass_2_id_attr.keys()).index(key)
        if curr_attr_to_set_idx != key_idx:
            raise ValueError(
                f"{key} is not yet to be defined. Currently, "
                f"{list(dataclass_2_id_attr.keys())[curr_attr_to_set_idx]} is the next to be set.")
        key_attr_name: str = dataclass_2_id_attr[key]
        setattr(self, key_attr_name, value)

    def peek(self) -> ScopedLocationAttr:
        """
        Pop without removing the last element: return the last non-None element.
        """
        return self.as_list()[-1]

    def pop(self) -> ScopedLocationAttr:
        """
        Returns the last `not None` scoped location attribute and remove it from self.
        """
        for attr_name in reversed(dataclass_2_id_attr.values()):
            attr_value = getattr(self, attr_name)
            if attr_value is not None:
                setattr(self, attr_name, None)
                return attr_value

    def popped(self) -> 'ScopedLocation':
        """
        Returns a new :class:`ScopedLocation` with the last `not None` scoped location attribute removed.
        """
        return ScopedLocation(*self.as_list()[:-1])

    def __len__(self) -> int:
        """
        Returns the number of not none scoped location attributes.
        """
        return len(self.as_list())

    def as_list(self, stringify: bool = False) -> List[ScopedLocationAttr]:
        """
        Returns a list of the scoped location attributes.
        Item that has a value of None are not included in the list.

        :param stringify: If True, the scoped location attributes are converted to strings.
        :return: list of not none scoped location attributes
        """
        _list: List[ScopedLocationAttr] = [
            getattr(self, attr_name)
            for attr_name in dataclass_2_id_attr.values()
        ]
        _list_ret = []
        for i in _list + [None]:
            if i is not None:
                if stringify:
                    i = str(i)
                _list_ret.append(i)
            else:
                break
        return _list_ret

    def new_dataclass_from_id(self):
        """
        Creates a new :class:`BaseDataclass` of the right type with
        just the provided ID filled.
        """
        dataklass = self.last_dc_type()

        return dataklass().set_id(self.as_list()[-1])

    def last_dc_type(self) -> Type['BaseDataclass']:
        dataklass: Type[BaseDataclass] = list(dataclass_2_id_attr.keys())[len(self) - 1]
        return dataklass

    def __eq__(self, other: 'ScopedLocation') -> bool:
        return self.as_list() == other.as_list()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.as_list())[1:-1]})"

    def __str__(self) -> str:
        return self.__repr__()


@dataclass(order=True)
class BaseDataclass(Generic[SubDataclassT], ABC):

    def __post_init__(self):
        self._validate()

    def _validate(self):
        for k, v in zip(self.get_sublocation_keys(), self.get_sublocation_values()):
            if isinstance(v, dict) and '__type__' in v:
                assert v['__type__'] == dataclass_2_subdataclass[self.__class__].__name__
                v: SubDataclassT = BaseDataclass.from_dict(v)
                assert v.get_id() == k
                self.store(v)

        subdataklass: Type[SubDataclassT] = dataclass_2_subdataclass[self.__class__]
        if subdataklass is not None and not all(
            (isinstance(s, subdataklass) for s in self.get_sublocation_values() if s is not None)
        ):
            raise ValueError(f"{self.__class__.__name__} must have all sublocation values as {subdataklass.__name__}.")

    @staticmethod
    def from_dict(_dict: Dict[str, Any]) -> 'BaseDataclass':
        if '__type__' not in _dict:
            raise ValueError("Dict must have a __type__ key.")
        _dict: dict = copy.copy(_dict)
        _type_str: str = _dict.pop('__type__')
        _klass: type = str_2_dataclass[_type_str]

        return _klass(**_dict)

    def to_dict(self) -> Dict[str, Any]:
        _dict = OrderedDict()
        _dict['__type__'] = self.__class__.__name__
        for _field in fields(self):
            _attr = getattr(self, _field.name)
            if _attr is not None:
                if isinstance(_attr, BaseDataclass):
                    _attr = _attr.to_dict()
                _dict[_field.name] = _attr
        return _dict

    @property
    def _id_attr_name(self) -> str:
        return dataclass_2_id_attr[self.__class__]

    @property
    def _sublocation_attr_name(self) -> str:
        return dataclass_2_subloc_attr[self.__class__]

    def get_id(self) -> ScopedLocationAttr:
        return getattr(self, self._id_attr_name)

    def set_id(self, _id: ScopedLocationAttr) -> 'BaseDataclass':
        setattr(self, self._id_attr_name, _id)
        self._validate()
        return self

    def get_sublocation(self) -> Union[List[SubDataclassT], 'OrderedDict[str, SubDataclassT]']:
        return getattr(self, dataclass_2_subloc_attr[self.__class__])

    def set_sublocation(self, sublocation: Union[List[SubDataclassT], 'OrderedDict[str, SubDataclassT]']) -> 'BaseDataclass':
        setattr(self, dataclass_2_subloc_attr[self.__class__], sublocation)
        return self

    @abstractmethod
    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        """
        Use this to set a shallow sublocation only from their keys.
        """
        raise NotImplementedError("This is an abstract method.")

    def __contains__(self, key: ScopedLocation) -> bool:
        # Too shallow:
        if len(key) < list(dataclass_2_subdataclass.keys()).index(self.__class__):
            raise ValueError(f"Key not deep enough for this dataclass of type {self.__class__.__name__}")

        # Terminal recursion condition:
        if key.at_dc(self) == key:
            return key.peek() == self.get_id()

        # Recursive call deeper otherwise:
        key_subattr: ScopedLocationAttr = key[self.subdataclass_type()]
        if self._is_key_in_subattr(key_subattr):
            return key in self.get_sublocation()[key_subattr]
        return False

    def __getitem__(self, loc: ScopedLocation) -> 'SubDataclassT':
        subdataklass: Type[SubDataclassT] = dataclass_2_subdataclass[self.__class__]
        if subdataklass is None:
            return self
        located_sub_attr: ScopedLocationAttr = loc[subdataklass]
        if located_sub_attr is None:
            return self
        if not self._is_key_in_subattr(located_sub_attr):
            raise KeyError(f"Item at loc={loc} is not in {self.shallow()}.")
        subloc: SubDataclassT = self.get_sublocation()[located_sub_attr]
        sublocs_sublocs: SubDataclassT = subloc[loc]
        return sublocs_sublocs

    def _is_key_in_subattr(self, key_subattr):
        return key_subattr in self.get_sublocation_keys()

    def __len__(self) -> int:
        return len(self.get_sublocation())

    @abstractmethod
    def store(self, dc: SubDataclassT) -> ScopedLocationAttr:
        """
        Add a subdataclass to the sublocation, at its proper ID.
        """
        raise NotImplementedError("Must use mixins.")

    @abstractmethod
    def shallow(self) -> 'BaseDataclass':
        """
        Replaces the sublocation items with None when the sublocation is a BaseDataclass type.
        """
        raise NotImplementedError("Must use mixins.")

    @abstractmethod
    def empty(self) -> 'BaseDataclass':
        """
        Do empty the sublocation when the sublocation is a BaseDataclass type.
        """
        raise NotImplementedError("Must use mixins.")

    def has_sublocation_dataclasses(self) -> bool:
        return self.__class__ in list(dataclass_2_subdataclass.keys())[:-1]

    @abstractmethod
    def get_sublocation_values(self) -> List[SubDataclassT]:
        raise NotImplementedError("Must use mixins.")

    @abstractmethod
    def get_sublocation_keys(self) -> List[ScopedLocationAttr]:
        raise NotImplementedError("Must use mixins.")

    @classmethod
    def subdataclass_type(cls) -> Type[SubDataclassT]:
        return dataclass_2_subdataclass[cls]


@dataclass(order=True)
class DataclassHasOrderedDictMixin:

    def _validate(self):
        super()._validate()
        if not isinstance(self.get_sublocation(), OrderedDict):
            raise ValueError(f"{self.__class__.__name__} must have an OrderedDict as sublocation.")
        if not all(
            (isinstance(s, ScopedLocationAttrStr) for s in self.get_sublocation_keys())
        ):
            raise ValueError(f"{self.__class__.__name__} must have all sublocation keys as strings.")

    def get_sublocation(self) -> 'OrderedDict[ScopedLocationAttrStr, SubDataclassT]':
        ret = super().get_sublocation()
        return ret

    def set_sublocation(self, sublocation: 'OrderedDict[ScopedLocationAttrStr, SubDataclassT]') -> None:
        setattr(self, self._sublocation_attr_name, sublocation)
        self._validate()

    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        _sublocation = OrderedDict([(k, None) for k in keys])
        setattr(self, self._sublocation_attr_name, _sublocation)
        self._validate()

    def get_sublocation_values(self) -> List[SubDataclassT]:
        return list(self.get_sublocation().values())

    def get_sublocation_keys(self) -> List[ScopedLocationAttrStr]:
        return list(self.get_sublocation().keys())

    def store(self, dc: SubDataclassT) -> ScopedLocationAttrStr:
        self.get_sublocation()[dc.get_id()] = dc
        self._validate()
        return dc.get_id()

    def shallow(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        new_sublocation = OrderedDict([(k, None) for k in self.get_sublocation().keys()])
        self_copy.set_sublocation(new_sublocation)
        # Deep copy only after trimming sublocations.
        return copy.deepcopy(self_copy)

    def empty(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        self_copy.set_sublocation(OrderedDict())
        return copy.deepcopy(self_copy)


@dataclass(order=True)
class DataclassHasListMixin:

    def _validate(self):
        super()._validate()
        if not isinstance(self.get_sublocation(), List):
            raise ValueError(f"{self.__class__.__name__} must have a List as sublocation.")

    def get_sublocation(self) -> List[SubDataclassT]:
        return super().get_sublocation()

    def set_sublocation(self, sublocation: List[SubDataclassT]) -> None:
        setattr(self, self._sublocation_attr_name, sublocation)
        self._validate()

    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        _sublocation = [None for k in keys]
        assert (set(range(len(_sublocation))) == set(
            [int(k) for k in keys])), f"Bad sublocation keys are being set into DataclassHasListMixin (type {self.__class__.__name__}, id={self.get_id()}) : {keys}."
        setattr(self, self._sublocation_attr_name, _sublocation)
        self._validate()

    def get_sublocation_values(self) -> List[SubDataclassT]:
        return self.get_sublocation()

    def get_sublocation_keys(self) -> List[ScopedLocationAttrInt]:
        return list(range(len(self)))

    def _is_key_in_subattr(self, key_subattr):
        return key_subattr in self.get_sublocation_keys() or (key_subattr == -1 and len(self) != 0)

    def store(self, dc: SubDataclassT) -> ScopedLocationAttrInt:
        _id = dc.get_id()
        if _id == self.get_next_i():
            self.get_sublocation().append(dc)
        elif _id < self.get_next_i():
            self.get_sublocation()[dc.get_id()] = dc
        else:
            raise ValueError(f"{dc} has id {dc.get_id()} which is greater than the next id {self.get_next_i()}.")
        self._validate()
        return _id

    def shallow(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        new_sublocation = [None for _ in range(len(self))]
        self_copy.set_sublocation(new_sublocation)
        # Deep copy only after trimming sublocations.
        return copy.deepcopy(self_copy)

    def empty(self) -> 'BaseDataclass':
        self_copy = copy.copy(self)
        self_copy.set_sublocation([])
        return copy.deepcopy(self_copy)

    def get_next_i(self) -> ScopedLocationAttrInt:
        return len(self.get_sublocation())


@dataclass(order=True)
class BaseTrialDataclassMixin:
    """
    Mixin class for :class:`TrialMetadata` and :class:`TrialSplitMetadata` that
    also must inherit from :class:`BaseMetadata`.
    """
    hyperparams: HyperparameterSamples = field(default_factory=HyperparameterSamples)
    status: TrialStatus = TrialStatus.PLANNED
    created_time: datetime.datetime = field(default_factory=datetime.datetime.now)
    start_time: datetime.datetime = field(default_factory=datetime.datetime.now)
    end_time: datetime.datetime = None
    # error: str = None
    # error_traceback: str = None
    # logs: List[str] = field(default_factory=list)

    def _validate(self):
        super()._validate()
        if not isinstance(self.hyperparams, HyperparameterSamples):
            self.hyperparams = HyperparameterSamples(self.hyperparams)
        self._validate_date_attr("created_time")
        self._validate_date_attr("start_time")
        self._validate_date_attr("end_time")
        if not isinstance(self.status, TrialStatus):
            self.status = TrialStatus(self.status)

    def _validate_date_attr(self, attr_name) -> None:
        attr_value = getattr(self, attr_name)
        if attr_value is not None and not isinstance(attr_value, datetime.datetime):
            attr_value = datetime.datetime.strptime(attr_value, LOGGING_DATETIME_STR_FORMAT)
            setattr(self, attr_name, attr_value)

    def start(self) -> 'BaseTrialDataclassMixin':
        self.status = TrialStatus.RUNNING
        self.start_time = datetime.datetime.now()
        self._validate()
        return self

    def end(self, status: TrialStatus) -> 'BaseTrialDataclassMixin':
        self.status = status
        self.end_time = datetime.datetime.now()
        self._validate()
        return self


@dataclass(order=True)
class RootDataclass(DataclassHasOrderedDictMixin, BaseDataclass['ProjectDataclass']):
    projects: typing.OrderedDict[str, 'ProjectDataclass'] = field(
        default_factory=lambda: OrderedDict({DEFAULT_PROJECT: ProjectDataclass()}))

    @property
    def _id_attr_name(self) -> str:
        return None

    @property
    def _sublocation_attr_name(self) -> str:
        return "projects"

    def get_id(self) -> ScopedLocationAttr:
        return None

    def get_sublocation(self) -> 'OrderedDict[str, SubDataclassT]':
        return self.projects


@dataclass(order=True)
class ProjectDataclass(DataclassHasOrderedDictMixin, BaseDataclass['ClientDataclass']):
    project_name: str = DEFAULT_PROJECT
    clients: typing.OrderedDict[str, 'ClientDataclass'] = field(
        default_factory=lambda: OrderedDict({DEFAULT_CLIENT: ClientDataclass()}))


@dataclass(order=True)
class ClientDataclass(DataclassHasListMixin, BaseDataclass['RoundDataclass']):
    client_name: ScopedLocationAttrStr = DEFAULT_CLIENT
    rounds: List['RoundDataclass'] = field(default_factory=list)


@dataclass(order=True)
class RoundDataclass(DataclassHasListMixin, BaseDataclass['TrialDataclass']):
    round_number: ScopedLocationAttrInt = 0
    trials: List['TrialDataclass'] = field(default_factory=list)
    main_metric_name: str = None


@dataclass(order=True)
class TrialDataclass(DataclassHasListMixin, BaseTrialDataclassMixin, BaseDataclass['TrialSplitDataclass']):
    """
    This class is a data structure most often used under :class:`AutoML` to store information about a trial.
    This information is itself managed by the :class:`HyperparameterRepository` class
    and the :class:`Trial` class within the AutoML.
    """
    trial_number: ScopedLocationAttrInt = 0
    validation_splits: List['TrialSplitDataclass'] = field(default_factory=list)
    retrained_split: 'TrialSplitDataclass' = None

    def store(self, dc: SubDataclassT) -> ScopedLocationAttrInt:
        if dc.get_id() == RETRAIN_TRIAL_SPLIT_ID:
            self.retrained_split = dc
            return RETRAIN_TRIAL_SPLIT_ID
        else:
            return super().store(dc)

    def __contains__(self, key: ScopedLocation) -> bool:
        if key[self.__class__] != self.get_id():
            return False
        if key[self.subdataclass_type()] == RETRAIN_TRIAL_SPLIT_ID:
            if self.retrained_split is None:
                return False
            return key in self.retrained_split
        else:
            return super().__contains__(key)

    def __getitem__(self, loc: ScopedLocation) -> SubDataclassT:
        if loc[self.__class__] != self.get_id():
            raise KeyError(f"Item at loc={loc} is not in {self.shallow()}.")
        if loc[self.subdataclass_type()] == RETRAIN_TRIAL_SPLIT_ID:
            if self.retrained_split is None:
                raise KeyError(f"Item at loc={loc} is not in {self.shallow()}. No retrains are available.")
            return self.retrained_split[loc]
        else:
            return super().__getitem__(loc)

    def set_sublocation_keys(self, keys: List[ScopedLocationAttr]) -> 'BaseDataclass':
        keys = [k for k in keys if int(k) != RETRAIN_TRIAL_SPLIT_ID]
        super().set_sublocation_keys(keys)

    def shallow(self) -> 'BaseDataclass':
        shallowed: TrialDataclass = super().shallow()
        if shallowed.retrained_split is not None and not isinstance(shallowed.retrained_split, str):
            shallowed.retrained_split = (
                f"ShallowedRetrainedSplitWithId({shallowed.retrained_split.get_id()})")
        return shallowed


RETRAIN_TRIAL_SPLIT_ID = -1


@dataclass(order=True)
class TrialSplitDataclass(DataclassHasOrderedDictMixin, BaseTrialDataclassMixin, BaseDataclass['MetricResultsDataclass']):
    """
    TrialSplit object used by AutoML algorithm classes.
    """
    split_number: ScopedLocationAttrInt = 0
    metric_results: typing.OrderedDict[str, 'MetricResultsDataclass'] = field(default_factory=OrderedDict)
    # introspection_data: RecursiveDict[str, Number] = field(default_factory=RecursiveDict)

    def is_retrain_split(self) -> bool:
        return self.split_number == -1


@dataclass(order=True)
class MetricResultsDataclass(DataclassHasListMixin, BaseDataclass[float]):
    """
    MetricResult object used by AutoML algorithm classes.
    """
    metric_name: ScopedLocationAttrStr = "main"
    validation_values: List[float] = field(default_factory=list)  # one per epoch.
    train_values: List[float] = field(default_factory=list)
    higher_score_is_better: bool = True

    def shallow(self) -> 'BaseDataclass':
        return copy.deepcopy(self)

    def empty(self) -> 'BaseDataclass':
        return copy.deepcopy(self)


dataclass_2_id_attr: typing.OrderedDict[BaseDataclass, str] = OrderedDict([
    (ProjectDataclass, "project_name"),
    (ClientDataclass, "client_name"),
    (RoundDataclass, "round_number"),
    (TrialDataclass, "trial_number"),
    (TrialSplitDataclass, "split_number"),
    (MetricResultsDataclass, "metric_name"),
])

dataclass_2_subloc_attr: typing.OrderedDict[BaseDataclass, str] = OrderedDict([
    (RootDataclass, "projects"),
    (ProjectDataclass, "clients"),
    (ClientDataclass, "rounds"),
    (RoundDataclass, "trials"),
    (TrialDataclass, "validation_splits"),
    (TrialSplitDataclass, "metric_results"),
    (MetricResultsDataclass, "validation_values"),
])

dataclass_2_subdataclass: typing.OrderedDict[BaseDataclass, SubDataclassT] = OrderedDict([
    (RootDataclass, ProjectDataclass),
    (ProjectDataclass, ClientDataclass),
    (ClientDataclass, RoundDataclass),
    (RoundDataclass, TrialDataclass),
    (TrialDataclass, TrialSplitDataclass),
    (TrialSplitDataclass, MetricResultsDataclass),
    (MetricResultsDataclass, None),
])

str_2_dataclass: typing.OrderedDict[str, BaseDataclass] = OrderedDict([
    (RootDataclass.__name__, RootDataclass),
    (ProjectDataclass.__name__, ProjectDataclass),
    (ClientDataclass.__name__, ClientDataclass),
    (RoundDataclass.__name__, RoundDataclass),
    (TrialDataclass.__name__, TrialDataclass),
    (TrialSplitDataclass.__name__, TrialSplitDataclass),
    (MetricResultsDataclass.__name__, MetricResultsDataclass),
    (RecursiveDict.__name__, RecursiveDict),
])


def as_named_odict(
    obj: Union[BaseDataclass, List[BaseDataclass]]
) -> 'OrderedDict[ScopedLocationAttrStr, BaseDataclass]':

    if isinstance(obj, BaseDataclass):
        obj = [obj]

    return OrderedDict([
        (i.get_id(), i) for i in obj
    ])


def object_pairs_decoder(obj):
    return object_decoder(obj, odictify=True)


def object_decoder(obj, odictify=False):
    if odictify:
        obj = OrderedDict(obj)
    if '__type__' in obj and obj['__type__'] in str_2_dataclass:
        # cls: Type = str_2_dataclass[obj['__type__']]
        return BaseDataclass.from_dict(obj)
        # kwargs = dict(obj)
        # del kwargs['__type__']
        # return cls(**kwargs)
    return obj


class MetadataJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, TrialStatus):
            return o.value
        if isinstance(o, RecursiveDict):
            return o.to_flat_dict()
        if isinstance(o, datetime.datetime):
            return o.strftime(LOGGING_DATETIME_STR_FORMAT)
        if isinstance(o, BaseDataclass):
            return o.to_dict()
        return JSONEncoder.encode(self, o)


def to_json(obj: BaseDataclass) -> str:
    return json.dumps(obj, cls=MetadataJSONEncoder)


def from_json(_json: str) -> BaseDataclass:
    return json.loads(_json, object_pairs_hook=object_pairs_decoder, object_hook=object_decoder)


class HyperparamsRepository(_ObservableRepo[Tuple['HyperparamsRepository', BaseDataclass]], BaseService):
    """
    Hyperparams repository that saves hyperparams, and scores for every AutoML trial.
    Cache folder can be changed to do different round numbers.

    For more information, read this `article by Martin Fowler on DDD Aggregates <https://martinfowler.com/bliki/DDD_Aggregate.html>`_.

    .. seealso::
        :class:`AutoML`,
        :class:`Trainer`,
    """

    def __init__(self):
        BaseService.__init__(self)
        _ObservableRepo.__init__(self)

    @abstractmethod
    def load(self, scope: ScopedLocation, deep=False) -> SubDataclassT:
        """
        Get metadata from scope.

        The fetched metadata will be the one that is the last item
        that is not a None in the provided scope.

        :param scope: scope to get metadata from.
        :return: metadata from scope.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")

    @abstractmethod
    def save(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False) -> 'HyperparamsRepository':
        """
        Save metadata to scope.

        :param metadata: metadata to save.
        :param scope: scope to save metadata to.
        :param deep: if True, save metadata's sublocations recursively.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")

    @abstractmethod
    def add_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> 'HyperparamsRepository':
        """
        Add logging handler to repository's provided scope.
        Handler must be set manually for each scope below this scope.

        :param logger: logger to add handler to.
        :param scope: scope to add handler to.
        :return: self.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")

    @abstractmethod
    def get_log_from_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> str:
        """
        Get log from repository's provided scope and handler that was set with :func:`add_logging_handler`.

        :param scope: scope to get log from.
        :return: log from scope.
        """
        raise NotImplementedError("Use a concrete class. This is an abstract class.")


class _InMemoryRepositoryLoggerHandlerMixin:
    """
    Mixin to add a in-memory logging handler to a repository.
    """

    def add_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> 'HyperparamsRepository':
        return self

    def get_log_from_logging_handler(self, logger: NeuraxleLogger, scope: ScopedLocation) -> str:
        return logger.get_scoped_string_history()


class VanillaHyperparamsRepository(_InMemoryRepositoryLoggerHandlerMixin, HyperparamsRepository):
    """
    Hyperparams repository that saves data AutoML-related info.
    """

    def __init__(
        self,
        cache_folder: str
    ):
        """
        :param cache_folder: folder to store trials.
        :param hyperparams_repo_class: class to use to save hyperparams.
        :param hyperparams_repo_kwargs: kwargs to pass to hyperparams_repo_class.
        """
        HyperparamsRepository.__init__(self)
        _InMemoryRepositoryLoggerHandlerMixin.__init__(self)
        self.cache_folder = os.path.join(cache_folder, self.__class__.__name__)
        self.root: RootDataclass = RootDataclass()

    @staticmethod
    def from_root(root: RootDataclass, cache_folder: str) -> 'VanillaHyperparamsRepository':

        return VanillaHyperparamsRepository(
            cache_folder=cache_folder,
        ).save(root, ScopedLocation(), deep=True)

    def load(self, scope: ScopedLocation, deep=False) -> SubDataclassT:
        try:
            ret: BaseDataclass = self.root[scope]
            if not deep:
                ret = ret.shallow()
        except KeyError:
            ret: BaseDataclass = scope.new_dataclass_from_id()

        return copy.deepcopy(ret)

    def save(self, _dataclass: SubDataclassT, scope: ScopedLocation, deep=False) -> 'VanillaHyperparamsRepository':
        # Sanitizing
        _dataclass: SubDataclassT = copy.deepcopy(_dataclass)
        scope = scope.at_dc(_dataclass)

        # Sanity checks: good type
        if not isinstance(_dataclass, BaseDataclass):
            raise ValueError(f"Was expecting a dataclass. Got `{_dataclass.__class__.__name__}`.")
        # Sanity checks: sufficient scope depth
        scope = scope[:_dataclass.__class__]  # Sanitizing scope to dtype loc.
        _id_from_scope: ScopedLocationAttr = scope[_dataclass.__class__]
        if _id_from_scope is not None:
            if _id_from_scope != _dataclass.get_id():
                raise ValueError(
                    f"The scope `{scope}` with {_dataclass.__class__.__name__} id `{_id_from_scope}` does not match the provided dataclass id `{_dataclass.get_id()}` for `{_dataclass}`."
                )
            scope.pop()
            # else check if the scope is at least of the good class length:
        elif len(scope) != list(dataclass_2_subloc_attr.keys()).index(_dataclass.__class__):
            raise ValueError(
                f"The scope `{scope}` is not of the good length for dataclass type `{_dataclass.__class__.__name__}`."
            )

        # Passthrough dc's sublocation when saving shallow:
        if not deep:
            if isinstance(_dataclass, RootDataclass):
                _dataclass.set_sublocation(self.root.get_sublocation())
            elif scope.with_dc(_dataclass) in self.root:
                prev_dc: SubDataclassT = self.root[scope.with_dc(_dataclass)]
                _dataclass.set_sublocation(prev_dc.get_sublocation())
            else:
                _dataclass = _dataclass.empty()

        # Finally storing.
        if isinstance(_dataclass, RootDataclass):
            self.root = _dataclass
        else:
            self.root[scope].store(_dataclass)
        return self


class InMemoryHyperparamsRepository(RaiseDeprecatedClass, _InMemoryRepositoryLoggerHandlerMixin, HyperparamsRepository):
    """
    In memory hyperparams repository that can print information about trials.
    Useful for debugging.
    """

    def __init__(self, *kargs, **kwargs):
        RaiseDeprecatedClass.__init__(
            self,
            replacement_class=VanillaHyperparamsRepository,
            since_version="0.7.0",
        )


class BaseHyperparameterOptimizer(ABC):

    @abstractmethod
    def find_next_best_hyperparams(self, round_scope) -> HyperparameterSamples:
        """
        Find the next best hyperparams using previous trials, that is the
        whole :class:`neuraxle.metaopt.data.aggregate.Round`.

        :param round: a :class:`neuraxle.metaopt.data.aggregate.Round`
        :return: next hyperparameter samples to train on
        """
        raise NotImplementedError()


class HyperparameterSamplerStub(BaseHyperparameterOptimizer):

    def __init__(self, preconfigured_hp_samples: HyperparameterSamples):
        self.preconfigured_hp_samples = preconfigured_hp_samples

    def find_next_best_hyperparams(self, round_scope) -> HyperparameterSamples:
        return self.preconfigured_hp_samples


class AutoMLContext(CX):

    @property
    def logger(self) -> NeuraxleLogger:
        self.add_scoped_logger_file_handler()
        return CX.logger.fget(self)

    @property
    def logger_at_scoped_loc(self) -> NeuraxleLogger:
        return logging.getLogger(self.get_identifier(include_step_names=False))

    def add_scoped_logger_file_handler(self) -> NeuraxleLogger:
        """
        Add a file handler to the logger at the current scoped location to capture logs
        at this scope and below this scope.
        """
        self.repo.add_logging_handler(self.logger_at_scoped_loc, self.loc)

    def free_scoped_logger_file_handler(self):
        """
        Remove file handlers from logger to free file lock (especially on Windows).
        """
        self.logger_at_scoped_loc.without_file_handler()

    def read_scoped_log(self) -> str:
        """
        Read the scoped logger file.
        """
        return self.repo.get_log_from_logging_handler(self.logger, self.loc)

    def copy(self):
        copy_kwargs = self._get_copy_kwargs()
        return AutoMLContext(**copy_kwargs)

    def _get_copy_kwargs(self):
        return super()._get_copy_kwargs()

    @staticmethod
    def from_context(
        context: CX = None,
        repo: HyperparamsRepository = None,
        loc: ScopedLocation = None
    ) -> 'AutoMLContext':
        """
        Create a new AutoMLContext from an ExecutionContext.

        :param context: ExecutionContext
        """
        new_context: AutoMLContext = AutoMLContext.copy(
            context if context is not None else AutoMLContext()
        )
        if not new_context.has_service(HyperparamsRepository):
            new_context.register_service(
                HyperparamsRepository,
                repo or VanillaHyperparamsRepository(new_context.get_path())
            )
        if not new_context.has_service(ScopedLocation):
            new_context.register_service(
                ScopedLocation,
                loc or ScopedLocation()
            )
        if not new_context.has_service(ContextLock):
            new_context.register_service(
                ContextLock,
                ContextLock()
            )
        return new_context

    @property
    def lock(self):
        return self.synchroneous()

    @property
    def loc(self) -> ScopedLocation:
        return self.get_service(ScopedLocation)

    @property
    def repo(self) -> HyperparamsRepository:
        return self.get_service(HyperparamsRepository)

    def push_attr(self, subdataclass: SubDataclassT) -> 'AutoMLContext':
        """
        Push a new attribute into the ScopedLocation.

        :param name: attribute name
        :param value: attribute value
        :return: an AutoMLContext copy with the new loc attribute.
        """
        if not isinstance(subdataclass, BaseDataclass) and isinstance(subdataclass, (str, int)):
            return self.with_loc(self.loc.with_id(subdataclass))  # ID instead.
        return self.with_loc(self.loc.with_dc(subdataclass))

    def pop_attr(self) -> 'AutoMLContext':
        """
        Pop an attribute from the ScopedLocation.

        :return: an AutoMLContext copy with the new popped loc attribute.
        """
        return self.with_loc(self.loc.popped())

    def with_loc(self, loc: ScopedLocation) -> 'AutoMLContext':
        """
        Replace the ScopedLocation by the one provided.

        :param loc: ScopedLocation
        :return: an AutoMLContext copy with the new loc attribute.
        """
        new_self: AutoMLContext = self.copy()
        new_self.register_service(ScopedLocation, loc)
        return new_self

    def load_dc(self, deep=True) -> BaseDataclass:
        """
        Load the current dc from the repo.
        """
        return self.repo.load(self.loc, deep)

    def load_agg(self, deep=True) -> 'agg.BaseAggregate':
        """
        Load the current agg from the repo.
        """
        return agg.BaseAggregate.from_context(self, is_deep=deep)



import pytest

from neuraxle.hyperparams.space import RecursiveDict, HyperparameterSamples

POINT_SEPARATOR = '.'


@pytest.mark.parametrize("separator", ["__", ".", "___"])
def test_recursive_dict_to_flat(separator):
    dict_values = {
        'hp': 1,
        'stepa': {
            'hp': 2,
            'stepb': {
                'hp': 3
            }
        }
    }
    r = RecursiveDict(separator=separator, **dict_values)

    r = r.to_flat_dict()

    expected_dict_values = {
        'hp': 1,
        'stepa{}hp'.format(separator): 2,
        'stepa{0}stepb{0}hp'.format(separator): 3
    }
    assert r == expected_dict_values


def test_recursive_dict_to_flat_different_separator():
    dict_values = {
        'hp': 1,
        'stepa': {
            'hp': 2,
            'stepb': {
                'hp': 3
            }
        }
    }
    r = RecursiveDict(separator='__', **dict_values)
    r['stepa'] = RecursiveDict(r['stepa'], separator='.')
    r['stepa']['stepb'] = RecursiveDict(r['stepa']['stepb'], separator='$$$')

    nested_r = r.to_nested_dict()
    r = r.to_flat_dict()

    expected_dict_values = {
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb.hp': 3
    }
    assert r == expected_dict_values
    assert nested_r == dict_values

def test_recursive_dict_to_nested_dict_constructor():
    dict_values = {
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb__hp': 3
    }
    r = HyperparameterSamples(**dict_values)

    expected_dict_values = {
        'hp': 1,
        'stepa': {
            'hp': 2,
            'stepb': {
                'hp': 3
            }
        }
    }
    assert r == HyperparameterSamples(**expected_dict_values)
    assert r.to_nested_dict() == expected_dict_values


def test_recursive_dict_get_item():
    dict_values = {
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb__hp': 3
    }
    r = HyperparameterSamples(**dict_values)

    assert r[None] == {'hp': 1}
    assert r["hp"] == 1
    assert r['stepa'].to_flat_dict() == {'hp': 2, 'stepb__hp': 3}
    assert r["stepa__hp"] == 2
    assert r["stepa"][None] == {'hp':2}
    assert r['stepa__stepb'].to_flat_dict() == {'hp': 3}
    assert r['stepa__stepb'][None] == {'hp': 3}
    assert r['stepa__stepb__hp'] == 3


def test_hyperparams_to_nested_dict_constructor():
    dict_values = {
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb__hp': 3
    }
    r = HyperparameterSamples(dict_values)

    expected_dict_values = {
        'hp': 1,
        'stepa': {
            'hp': 2,
            'stepb': {
                'hp': 3
            }
        }
    }
    assert r.to_nested_dict() == expected_dict_values
    assert r == HyperparameterSamples(expected_dict_values)


def test_recursive_dict_copy_constructor():
    dict_values = {
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb__hp': 3
    }
    r = RecursiveDict(RecursiveDict(**dict_values), separator='__')

    assert r == RecursiveDict(**dict_values)


def test_recursive_dict_copy_constructor_should_set_separator():
    dict_values = {
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb__hp': 3
    }
    r = RecursiveDict(RecursiveDict(**dict_values, separator=POINT_SEPARATOR))

    assert r.separator == POINT_SEPARATOR


def test_recursive_dict_should_raise_when_item_missing():
    with pytest.raises(KeyError):
        r = RecursiveDict()
        missing = r['missing']


@pytest.mark.parametrize("dict_values",
     [{
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb__hp': 3
    },{
        "stepa__hp1":1,
        'stepa__hp2': 2,
        'stepa__stepb__hp': 3
    }])
def test_hyperparams_copy_constructor(dict_values):
    r = HyperparameterSamples(HyperparameterSamples(**dict_values))
    assert r == HyperparameterSamples(**dict_values)


def test_hyperparams_to_flat():
    dict_values = {
        'hp': 1,
        'stepa': {
            'hp': 2,
            'stepb': {
                'hp': 3
            }
        }
    }
    r = HyperparameterSamples(**dict_values)

    r = r.to_flat_dict()

    expected_dict_values = {
        'hp': 1,
        'stepa__hp': 2,
        'stepa__stepb__hp': 3
    }
    assert r == expected_dict_values



"""
Streaming Pipelines for Parallel and Queued Data Processing
===================================================================

Neuraxle steps for streaming data in parallel in the pipeline.

Pipelines can stream data in queues with workers for each steps.
Max queue sizes can be set, as well as number of clones per steps
for the transformers.


..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
import time
import warnings
from abc import abstractmethod
from multiprocessing import Lock, Process, Queue, RLock
from threading import Thread
from typing import Any, Dict, Iterable, List, Tuple, Union

from neuraxle.base import BaseSaver, BaseTransformer
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import (MetaStep, MixinForBaseTransformer, NamedStepsList,
                           NonFittableMixin, _FittableStep)
from neuraxle.data_container import (DACT, AbsentValuesNullObject,
                                     ListDataContainer)
from neuraxle.hyperparams.space import RecursiveDict
from neuraxle.pipeline import Joiner, MiniBatchSequentialPipeline, Pipeline
from neuraxle.steps.numpy import NumpyConcatenateOuterBatch


class ObservableQueueMixin(MixinForBaseTransformer):
    """
    A class to represent a step that can put items in a queue so that the
    step can be used to consume its tasks on its own.

    Once tasks are solved, their results are added to the
    subscribers that were subscribed with the notify call.

    A subscriber is itself an ObservableQueueMixin and will thus be
    used to consume the tasks added to it in probably another thread.

    .. seealso::
        :class:`BaseStep`,
        :class:`QueuedPipelineTask`,
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`SequentialQueuedPipeline`
    """

    def __init__(self, queue: Queue):
        MixinForBaseTransformer.__init__(self)
        self.queue = queue
        self.observers = []
        self._add_observable_queue_step_saver()

    def _teardown(self):
        self.queue = None
        return self

    def _add_observable_queue_step_saver(self):
        if not hasattr(self, 'savers'):
            warnings.warn(
                'Please initialize Mixins in the good order. ObservableQueueMixin should be initialized after '
                'Appending the ObservableQueueStepSaver to the savers. Saving might fail.'
            )
            self.savers = [ObservableQueueStepSaver()]
        else:
            self.savers.append(ObservableQueueStepSaver())

    def subscribe_step(self, observer_queue_worker: 'ObservableQueueMixin') -> 'ObservableQueueMixin':
        """
        Subscribe a queue worker to self such that we can notify them to post tasks to put on them.
        The subscribed queue workers get notified when :func:`~neuraxle.distributed.streaming.ObservableQueueMixin.notify_step` is called.
        """
        self.observers.append(observer_queue_worker.queue)
        return self

    def get_task(self) -> 'QueuedPipelineTask':
        """
        Get last item in queue.
        """
        return self.queue.get()

    def put_task(self, value: DACT):
        """
        Put a queued pipeline task in queue.
        """
        self.queue.put(QueuedPipelineTask(step_name=self.name, data_container=value.copy()))

    def notify_step(self, value: DACT):
        """
        Notify all subscribed queue workers to put them some tasks on their queue.
        """
        for observer_queue in self.observers:
            observer_queue.put(QueuedPipelineTask(step_name=self.name, data_container=value))


class QueuedPipelineTask(object):
    """
    Data object to contain the tasks processed by the queued pipeline.
    Attributes: step_name, data_container

    .. seealso::
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`SequentialQueuedPipeline`
    """

    def __init__(self, data_container, step_name=None):
        self.step_name = step_name
        self.data_container = data_container


class ObservableQueueStepSaver(BaseSaver):
    """
    Saver for observable queue steps.

    .. seealso::
        :class:`QueueWorker`,
        :class:`neuraxle.base.BaseSaver`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`SequentialQueuedPipeline`
    """

    def save_step(self, step: BaseTransformer, context: 'CX') -> BaseTransformer:
        step.queue = None
        step.observers = []
        return step

    def can_load(self, step: BaseTransformer, context: 'CX') -> bool:
        return True

    def load_step(self, step: 'BaseTransformer', context: 'CX') -> 'BaseTransformer':
        step.queue = Queue()
        return step


class QueueWorker(ObservableQueueMixin, MetaStep):
    """
    Start multiple Process or Thread that process items from the queue of batches to process.
    It is both an observable, and observer.
    It notifies the results of the wrapped step handle transform method.
    It receives the next data container to process.

    .. seealso::
        :class:`Observer`,
        :class:`Observable`,
        :class:`MetaStepMixin`,
        :class:`BaseStep`
    """

    def __init__(
            self,
            wrapped: BaseTransformer,
            max_queue_size: int,
            n_workers: int,
            use_processes: bool = True,
            additional_worker_arguments: List = None,
            use_savers: bool = False
    ):
        if not additional_worker_arguments:
            additional_worker_arguments = [[] for _ in range(n_workers)]

        MetaStep.__init__(self, wrapped)
        ObservableQueueMixin.__init__(self, Queue(maxsize=max_queue_size))  # max_queue_size is in batches

        self.use_processes: bool = use_processes
        self.workers: List[Process] = []
        self.n_workers: int = n_workers
        self.observers: List[Queue] = []
        self.additional_worker_arguments = additional_worker_arguments
        self.use_savers = use_savers

    def __getstate__(self):
        """
        This class, upon being forked() to a new process with pickles,
        should not copy references to other threads.
        """
        state = self.__dict__.copy()
        state['workers'] = None
        return state

    def start(self, context: CX):
        """
        Start multiple processes or threads with the worker function as a target.

        :param context: execution context
        :type context: ExecutionContext
        :return:
        """
        if self.use_savers:
            _ = self.save(context, full_dump=True)  # Cannot delete queue worker self.
            del self.wrapped  # queue is deleted

        thread_safe_context = context
        thread_safe_lock: RLock = context.synchroneous()
        parallel_call = Thread
        if self.use_processes:
            # New process requires trimming the references to other processes
            # when we create many processes: https://stackoverflow.com/a/65749012
            thread_safe_lock, thread_safe_context = context.thread_safe()
            parallel_call = Process

        self.workers = []
        for _, worker_arguments in zip(range(self.n_workers), self.additional_worker_arguments):
            p = parallel_call(
                target=worker_function,
                args=(self, thread_safe_lock, thread_safe_context, self.use_savers, worker_arguments)
            )
            p.daemon = True
            p.start()
            self.workers.append(p)

    def _teardown(self):
        """
        Stop all processes on teardown.

        :return: teardowned self
        """
        self.stop()
        return self

    def stop(self):
        """
        Stop all of the workers.

        :return:
        """
        if self.use_processes:
            [w.terminate() for w in self.workers]

        self.workers = []
        self.observers = []


def worker_function(queue_worker: QueueWorker, shared_lock: Lock, context: CX, use_savers: bool,
                    additional_worker_arguments):
    """
    Worker function that transforms the items inside the queue of items to process.

    :param queue_worker: step to transform
    :param context: execution context
    :param use_savers: use savers
    :param additional_worker_arguments: any additional arguments that need to be passed to the workers
    :return:
    """

    try:
        context.restore_lock(shared_lock)
        if use_savers:
            saved_queue_worker: QueueWorker = context.load(queue_worker.get_name())
            queue_worker.set_step(saved_queue_worker.get_step())
        step = queue_worker.get_step()

        additional_worker_arguments = tuple(
            additional_worker_arguments[i: i + 2] for i in range(0, len(additional_worker_arguments), 2)
        )

        for argument_name, argument_value in additional_worker_arguments:
            step.__dict__.update({argument_name: argument_value})

        while True:
            try:
                task: QueuedPipelineTask = queue_worker.get_task()
                data_container = step.handle_transform(task.data_container, context)
                queue_worker.notify_step(data_container)
            except Exception as err:
                queue_worker.notify_step(err)
            finally:
                time.sleep(0.005)
    except Exception as err:
        queue_worker.notify_step(err)


QueuedPipelineStepsTuple = Union[
    BaseTransformer,  # step
    Tuple[int, BaseTransformer],  # (n_workers, step)
    Tuple[str, BaseTransformer],  # (step_name, step)
    Tuple[str, int, BaseTransformer],  # (step_name, n_workers, step)
    Tuple[str, int, int, BaseTransformer],  # (step_name, n_workers, max_queue_size, step)
    Tuple[str, int, List[Tuple], BaseTransformer],  # (step_name, n_workers, additional_worker_arguments, step)
    Tuple[str, int, List[Tuple], BaseTransformer]  # (step_name, n_workers, additional_worker_arguments, step)
]


class BaseQueuedPipeline(MiniBatchSequentialPipeline):
    """
    Sub class of :class:`Pipeline`.
    Transform data in many pipeline steps at once in parallel in the pipeline using multiprocessing Queues.

    Example usage :

    .. code-block:: python

        # Multiple ways of specifying the steps tuples exists to do various things:
        # step name, step
        p = SequentialQueuedPipeline([
            ('step_a', Identity()),
            ('step_b', Identity()),
        ], n_workers=1, batch_size=10, max_queue_size=10)

        # step name, number of workers, step
        p = SequentialQueuedPipeline([
            ('step_a', 1, Identity()),
            ('step_b', 1, Identity()),
        ], batch_size=10, max_queue_size=10)

        # step name, number of workers, and max size
        p = SequentialQueuedPipeline([
            ('step_a', 1, 10, Identity()),
            ('step_b', 1, 10, Identity()),
        ], batch_size=10)

        # step name, number of workers for each step, and additional argument for each worker
        p = SequentialQueuedPipeline([
            ('step_a', 1, [('host', 'host1'), ('host', 'host2')], 10, Identity())
        ], batch_size=10)

        # step name, number of workers for each step, additional argument for each worker, and max size
        p = SequentialQueuedPipeline([
            ('step_a', 1, [('host', 'host1'), ('host', 'host2')], 10, Identity())
        ], batch_size=10)

        # It's also possible to do parallel feature unions:
        n_workers = 4
        worker_arguments = [('hyperparams', HyperparameterSamples({'multiply_by': 2})) for _ in range(n_workers)]
        p = ParallelQueuedFeatureUnion([
            ('1', n_workers, worker_arguments, MultiplyByN()),
        ], batch_size=10, max_queue_size=5)
        outputs = p.transform(list(range(100)))

    :param steps: pipeline steps.
    :param batch_size: number of elements to combine into a single batch.
    :param n_workers_per_step: number of workers to spawn per step.
    :param max_queue_size: max number of batches inside the processing queue between the workers.
    :param data_joiner: transformer step to join streamed batches together at the end of the pipeline.
    :param use_processes: use processes instead of threads for parallel processing. multiprocessing.context.Process is used by default.
    :param use_savers: use savers to serialize steps for parallel processing. Recommended if using processes instead of threads.
    :param keep_incomplete_batch: (Optional.) A bool that indicates whether
    or not the last batch should be dropped in the case it has fewer than
    `batch_size` elements; the default behavior is to keep the smaller batch.
    :param default_value_data_inputs: expected_outputs default fill value
    for padding and values outside iteration range, or :class:`~neuraxle.data_container.DataContainer.AbsentValuesNullObject`
    to trim absent values from the batch
    :param default_value_expected_outputs: expected_outputs default fill value
    for padding and values outside iteration range, or :class:`~neuraxle.data_container.DataContainer.AbsentValuesNullObject`
    to trim absent values from the batch

    .. seealso::
        :class:`QueueWorker`,
        :class:`QueueJoiner`,
        :class:`CustomPipelineMixin`,
        :class:`Pipeline`
    """

    def __init__(
            self,
            steps: List[QueuedPipelineStepsTuple],
            batch_size: int,
            n_workers_per_step: int = None,
            max_queue_size: int = None,
            data_joiner: BaseTransformer = None,
            use_processes: bool = False,
            use_savers: bool = False,
            keep_incomplete_batch: bool = True,
            default_value_data_inputs: Union[Any, AbsentValuesNullObject] = None,
            default_value_expected_outputs: Union[Any, AbsentValuesNullObject] = None,
    ):
        if data_joiner is None:
            data_joiner = NumpyConcatenateOuterBatch()
        self.data_joiner = data_joiner
        self.max_queue_size = max_queue_size
        self.n_workers_per_step = n_workers_per_step
        self.use_processes = use_processes
        self.use_savers = use_savers

        self.batch_size: int = batch_size
        self.keep_incomplete_batch: bool = keep_incomplete_batch
        self.default_value_data_inputs: Union[Any, AbsentValuesNullObject] = default_value_data_inputs
        self.default_value_expected_outputs: Union[Any, AbsentValuesNullObject] = default_value_expected_outputs

        MiniBatchSequentialPipeline.__init__(
            self,
            steps=self._initialize_steps_as_tuple(steps),
            batch_size=batch_size,
            keep_incomplete_batch=keep_incomplete_batch,
            default_value_data_inputs=default_value_data_inputs,
            default_value_expected_outputs=default_value_expected_outputs
        )
        self._refresh_steps()

    def _initialize_steps_as_tuple(self, steps: NamedStepsList) -> NamedStepsList:
        """
        Wrap each step by a :class:`QueueWorker` to  allow data to flow in many pipeline steps at once in parallel.

        :param steps: (name, n_workers, step)
        :type steps: NameNWorkerStepTupleList
        :return: steps as tuple
        """
        steps_as_tuple: NamedStepsList = []
        for step in steps:
            queue_worker = self._create_queue_worker(step)
            steps_as_tuple.append((queue_worker.name, queue_worker))

        steps_as_tuple.append(('queue_joiner', QueueJoiner(batch_size=self.batch_size)))

        return steps_as_tuple

    def _create_queue_worker(self, step: QueuedPipelineStepsTuple):
        name, n_workers, additional_worker_arguments, max_queue_size, actual_step = self._get_step_params(step)

        return QueueWorker(
            actual_step,
            n_workers=n_workers,
            use_processes=self.use_processes,
            max_queue_size=max_queue_size,
            additional_worker_arguments=additional_worker_arguments,
            use_savers=self.use_savers
        ).set_name('QueueWorker{}'.format(name))

    def _get_step_params(self, step):
        """
        Return all params necessary to create the QueuedPipeline for the given step.

        :param step: tuple
        :type step: QueuedPipelineStepsTupleList

        :return: return name, n_workers, max_queue_size, actual_step
        :rtype: tuple(str, int, int, BaseStep)
        """
        if isinstance(step, BaseTransformer):
            actual_step = step
            name = step.name
            max_queue_size = self.max_queue_size
            n_workers = self.n_workers_per_step
            additional_arguments = []
        elif len(step) == 2:
            if isinstance(step[0], str):
                name, actual_step = step
                n_workers = self.n_workers_per_step
            else:
                n_workers, actual_step = step
                name = actual_step.name
            max_queue_size = self.max_queue_size
            additional_arguments = []
        elif len(step) == 3:
            name, n_workers, actual_step = step
            max_queue_size = self.max_queue_size
            additional_arguments = []
        elif len(step) == 4:
            if isinstance(step[2], Iterable):
                name, n_workers, additional_arguments, actual_step = step
                max_queue_size = self.max_queue_size
            else:
                name, n_workers, max_queue_size, actual_step = step
                additional_arguments = []
        elif len(step) == 5:
            name, n_workers, additional_arguments, max_queue_size, actual_step = step
        else:
            raise Exception('Invalid Queued Pipeline Steps Shape.')

        return name, n_workers, additional_arguments, max_queue_size, actual_step

    def _will_process(
        self, data_container: DACT, context: CX
    ) -> Tuple[DACT, CX]:
        """
        Setup streaming pipeline before any handler methods.

        :param data_container: data container
        :param context: execution context
        :return:
        """
        self._setup(context=context)
        return data_container.copy(), context

    def _setup(self, context: CX = None) -> 'BaseTransformer':
        """
        Connect the queued workers together so that the data can correctly flow through the pipeline.

        :param context: execution context
        :return: step
        :rtype: BaseStep
        """
        if not self.is_initialized:
            self.connect_queued_pipeline()
        super()._setup(context=context)
        return RecursiveDict()

    def fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple[Pipeline, DACT]:
        """
        Fit transform sequentially if any step is fittable. Otherwise transform in parallel.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return:
        """
        all_steps_are_not_fittable = True
        for _, step in self[:-1]:
            if isinstance(step.get_step(), _FittableStep) and not isinstance(step.get_step(), NonFittableMixin):
                all_steps_are_not_fittable = False

        if all_steps_are_not_fittable:
            data_container = self.transform_data_container(data_container, context)
            data_container = self._did_transform(data_container, context)
            return self, data_container

        self.is_invalidated = True

        return super().fit_transform_data_container(data_container, context)

    def transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        Transform data container

        :param data_container: data container to transform.
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: data container
        """
        joiner = self[-1]
        n_batches = self.get_n_batches(data_container)
        joiner.set_n_batches(n_batches)

        # start steps.
        context.synchroneous()
        for step in list(self.values())[:-1]:
            step.start(context)

        # send batches to input queues and label queue output expected summaries.
        for batch_i, data_container_batch in enumerate(data_container.minibatches(
            batch_size=self.batch_size,
            keep_incomplete_batch=self.keep_incomplete_batch,
            default_value_data_inputs=self.default_value_data_inputs,
            default_value_expected_outputs=self.default_value_expected_outputs
        )):
            self.send_batch_to_queued_pipeline(batch_index=batch_i, data_container=data_container_batch)

        # join output queues.
        data_container = joiner.join(original_data_container=data_container)
        return data_container

    def _did_transform(self, data_container: DACT, context: CX) -> DACT:
        """
        Stop all of the workers after transform. Also, join the data using self.data_joiner.

        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: data container
        :rtype: DataContainer
        """
        for name, step in self[:-1]:
            step.stop()

        return self.data_joiner.handle_transform(data_container, context)

    @abstractmethod
    def get_n_batches(self, data_container) -> int:
        """
        Get the total number of batches that the queue joiner is supposed to receive.

        :param data_container: data container to transform
        :type data_container: DataContainer
        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def connect_queued_pipeline(self):
        """
        Connect all the queued workers together so that the data can flow through each step.

        :return:
        """
        raise NotImplementedError()

    @abstractmethod
    def send_batch_to_queued_pipeline(self, batch_index: int, data_container: DACT):
        """
        Send batches to queued pipeline. It is blocking if there is no more space available in the multiprocessing queues.
        Workers might return batches in a different order, but the queue joiner will reorder them at the end.
        The queue joiner will use the summary ids to reorder all of the received batches.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        raise NotImplementedError()


class SequentialQueuedPipeline(BaseQueuedPipeline):
    """
    Using :class:`QueueWorker`, run all steps sequentially even if they are in separate processes or threads.
    This is a parallel pipeline that uses a queue to communicate between the steps, and which parallelizes the steps
    using many workers in different processes or threads.

    This pipeline is useful when the steps are independent of each other and can be run in parallel. This is especially
    the case when the steps are not fittable, such as inheriting from the :class:`~neuraxle.base.NonFittableMixin`.
    Otherwise, fitting may not be parallelized, although the steps can be run in parallel for the transformation.

    .. seealso::
        :class:`~neuraxle.pipeline.BasePipeline`,
        :func:`~neuraxle.data_container.DataContainer.minibatches`,
        :class:`~neuraxle.data_container.DataContainer.AbsentValuesNullObject`,
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`ParallelQueuedPipeline`,
        :class:`QueueJoiner`,
        :class:`Observer`,
        :class:`Observable`
    """

    def get_n_batches(self, data_container) -> int:
        """
        Get the number of batches to process.

        :param data_container: data container to transform
        :return: number of batches
        """
        return data_container.get_n_batches(
            batch_size=self.batch_size,
            keep_incomplete_batch=self.keep_incomplete_batch
        )

    def connect_queued_pipeline(self):
        """
        Sequentially connect of the queued workers.

        :return:
        """
        for i, (name, step) in enumerate(self[1:]):
            self[i].subscribe_step(step)

    def send_batch_to_queued_pipeline(self, batch_index: int, data_container: DACT):
        """
        Send batches to process to the first queued worker.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        self[-1].summaries.append(data_container.get_ids_summary())
        self[0].put_task(data_container)


class ParallelQueuedFeatureUnion(BaseQueuedPipeline):
    """
    Using :class:`QueueWorker`, run all steps in parallel using QueueWorkers.

    .. seealso::
        :class:`QueueWorker`,
        :class:`BaseQueuedPipeline`,
        :class:`SequentialQueuedPipeline`,
        :class:`QueueJoiner`,
        :class:`Observer`,
        :class:`Observable`
    """

    def get_n_batches(self, data_container):
        """
        Get the number of batches to process by the queue joiner.

        :return:
        """
        return data_container.get_n_batches(self.batch_size) * (len(self) - 1)

    def connect_queued_pipeline(self):
        """
        Connect the queue joiner to all of the queued workers to process data in parallel.

        :return:
        """
        for name, step in self[:-1]:
            step.subscribe_step(self[-1])

    def send_batch_to_queued_pipeline(self, batch_index: int, data_container: DACT):
        """
        Send batches to process to all of the queued workers.

        :param batch_index: batch index
        :param data_container: data container batch
        :return:
        """
        for name, step in self[:-1]:
            queue_joiner: QueueJoiner = self[-1]
            queue_joiner.summaries.append(data_container.get_ids_summary())
            step.put_task(data_container)


class QueueJoiner(ObservableQueueMixin, Joiner):
    """
    Observe the results of the queue worker of type :class:`QueueWorker`.
    Synchronize all of the workers together.

    .. seealso::
        :class:`QueuedPipeline`,
        :class:`Observer`,
        :class:`ListDataContainer`,
        :class:`DataContainer`
    """

    def __init__(self, batch_size, n_batches=None):
        self.n_batches_left_to_do = n_batches
        self.summaries: List[str] = []
        self.result = {}
        Joiner.__init__(self, batch_size=batch_size)
        ObservableQueueMixin.__init__(self, Queue())

    def _teardown(self) -> 'BaseTransformer':
        """
        Properly clean queue, summary ids, and results during teardown.

        :return: teardowned self
        """
        ObservableQueueMixin._teardown(self)
        Joiner._teardown(self)
        self.summaries = []
        self.result: Dict[str, ListDataContainer] = dict()
        return self

    def set_n_batches(self, n_batches):
        self.n_batches_left_to_do = n_batches

    def join(self, original_data_container: DACT) -> DACT:
        """
        Return the accumulated results received by the on next method of this observer.

        :return: transformed data container
        :rtype: DataContainer
        """
        while self.n_batches_left_to_do > 0:
            task: QueuedPipelineTask = self.get_task()
            self.n_batches_left_to_do -= 1
            step_name = task.step_name

            if step_name not in self.result:
                self.result[step_name] = ListDataContainer(
                    ids=[],
                    data_inputs=[],
                    expected_outputs=[]
                )

            self.result[step_name].append_data_container_in_data_inputs(task.data_container)

        data_containers = self._join_all_step_results()
        self.result = {}
        return original_data_container.set_data_inputs(data_containers)

    def _join_all_step_results(self):
        """
        Concatenate all resulting data containers together.

        :return:
        """
        results = []
        for step_name, data_containers in self.result.items():
            self._raise_exception_throwned_by_workers_if_needed(data_containers)
            step_results = self._join_step_results(data_containers)
            results.append(step_results)

        return results

    def _raise_exception_throwned_by_workers_if_needed(self, data_containers: ListDataContainer):
        for dc in data_containers.data_inputs:
            if isinstance(dc, Exception):
                # an exception has been throwned by the worker so reraise it here!
                exception = dc
                raise exception

    def _join_step_results(self, data_containers: ListDataContainer):
        # reorder results by ids of summary
        data_containers.data_inputs.sort(key=lambda dc: self.summaries.index(dc.get_ids_summary()))

        step_results = ListDataContainer.empty()
        for data_container in data_containers.data_inputs:
            step_results.concat(data_container)

        return step_results



"""
Tests for Flask.
============================================

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""

import numpy as np

from neuraxle.rest.flask import JSONDataResponseEncoder, JSONDataBodyDecoder, FlaskRestApiWrapper
from neuraxle.base import BaseTransformer


def setup_api():
    class Decoder(JSONDataBodyDecoder):
        """This is a custom JSON decoder class that precedes the pipeline's transformation."""

        def decode(self, data_inputs):
            """
            Transform a JSON list object into an np.array object.

            :param data_inputs: json object
            :return: np array for data inputs
            """
            return np.array(data_inputs)

    class Encoder(JSONDataResponseEncoder):
        """This is a custom JSON response encoder class for converting the pipeline's transformation outputs."""

        def encode(self, data_inputs) -> dict:
            """
            Convert predictions to a dict for creating a JSON Response object.

            :param data_inputs:
            :return:
            """
            return {
                'predictions': np.array(data_inputs).tolist()
            }

    class Multiplier(BaseTransformer):
        def transform(self, data_inputs):
            return 2 * data_inputs

    app = FlaskRestApiWrapper(
        json_decoder=Decoder(),
        wrapped=Multiplier(),
        json_encoder=Encoder()
    ).get_app()

    app.testing = True

    test_client = app.test_client()

    return test_client


def test_api_wrapper_works():
    test_client = setup_api()
    data_inputs = [
        [0, 1, 2],
        [3, 4, 5],
    ]

    json_response = test_client.get('/', json=data_inputs)

    predictions_np_arr = np.array(json_response.json["predictions"])
    expected_outputs = 2 * np.array(data_inputs)
    assert np.array_equal(predictions_np_arr, expected_outputs)



"""
Neuraxle's Observable Classes
===================================================================
Base observable classes, implementing the Observer pattern.
Some of them are used to track the evolution of the optimization process.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

"""
from abc import abstractmethod
from typing import Set, TypeVar, Generic


BaseDataclassT = TypeVar('BaseDataclassT')


class _ObservableRepo(Generic[BaseDataclassT]):
    """
    This class is used to implement the Observer design pattern.
    The _Observable class is a subject that is being observed by the _Observer class.
    The type T is the type of the value that will be observed.

    There are methods that the observer can define to send the notification:
    - notify_next(value: T),
    - notify_complete(value: T).

    The _Observable class is a generic class, so it can be used with any type T.

    It is possible to subscribe and unsubscribe observers.
    A subscription is a call to the subscribe method of the _Observable class.
    A unsubscription is a call to the unsubscribe method of the _Observable class.
    Thus, the subsibers receive the notifications of the _Observable class.

    A notification is a call to one of the notify_* methods of the _Observable class.
    A notification is a call to one of the update_* methods of the _Observer class.
    """

    def __init__(self):
        self._observers: Set[_ObserverOfRepo[BaseDataclassT]] = set()

    def subscribe(self, observer: '_ObserverOfRepo[BaseDataclassT]'):
        self._observers.add(observer)

    def unsubscribe(self, observer: '_ObserverOfRepo[BaseDataclassT]'):
        self._observers.discard(observer)

    def notify_next(self, value: BaseDataclassT):
        for observer in self._observers:
            observer.update_next(value)

    def notify_complete(self, value: BaseDataclassT):
        for observer in self._observers:
            observer.update_complete(value)


class _ObserverOfRepo(Generic[BaseDataclassT]):
    """
    This class is used to implement the Observer design pattern.
    The _Observer class is an observer that is being notified by the _Observable class.
    The type T is the type of the value that will be observed.

    Upon receiving a notification, the _Observer class can define these methods:
    - update_next(value: T),
    - update_complete(value: T)

    These methods are called by the _Observable class observing the observer.
    """
    @abstractmethod
    def update_next(self, value: BaseDataclassT):
        pass

    def update_complete(self, value: BaseDataclassT):
        pass



"""
Miscelaneous Pipeline Steps
====================================
You can find here misc. pipeline steps, for example, callbacks useful for debugging, and a step cloner.

..
    Copyright 2019, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

..
    Thanks to Umaneo Technologies Inc. for their contributions to this Machine Learning
    project, visit https://www.umaneo.com/ for more information on Umaneo Technologies Inc.

"""

import time
from abc import ABC
from typing import Any, Callable, List, Optional, Tuple

from neuraxle.base import BaseStep, BaseTransformer
from neuraxle.base import ExecutionContext as CX
from neuraxle.base import ForceHandleOnlyMixin, HandleOnlyMixin, MetaStep
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.space import RecursiveDict

VALUE_CACHING = 'value_caching'


class AssertFalseStep(HandleOnlyMixin, BaseStep):
    """
    Assert False upon _transform_data_container and _fit_data_container.
    """

    def __init__(self, message: str = "This step should not fit nor transform."):
        BaseStep.__init__(self)
        self.message: str = message

    def _transform_data_container(self, data_container, context):
        self._assert(False, self.message, context)

    def _fit_data_container(self, data_container, context):
        self._assert(False, self.message, context)


class BaseCallbackStep(BaseStep, ABC):
    """Base class for callback steps."""

    def __init__(self, callback_function, more_arguments: List = tuple(),
                 hyperparams=None, fit_callback_function=None, transform_function=None):
        """
        Create the callback step with a function and extra arguments to send to the function

        :param callback_function: The function that will be called on events.
        :param more_arguments: Extra arguments that will be sent to the callback after the processed data (optional).
        """
        BaseStep.__init__(self, hyperparams=hyperparams)
        self.transform_function = transform_function
        self.callback_function = callback_function
        self.fit_callback_function = fit_callback_function
        self.more_arguments = more_arguments

    def _fit_callback(self, data_inputs, expected_outputs):
        """
        Will call the self.fit_callback_function() with the data being processed and the extra arguments specified.
        It has no other effect.

        :param data_inputs: data inputs to fit
        :param expected_outputs: expected outputs to fit

        :return: self
        """
        self.fit_callback_function((data_inputs, expected_outputs), *self.more_arguments)

    def _callback(self, data):
        """
        Will call the self.callback_function() with the data being processed and the extra arguments specified.
        It has no other effect.

        :param data_inputs: the data to process
        :return: None
        """
        self.callback_function(data, *self.more_arguments)

    def fit(self, data_inputs, expected_outputs=None):
        return self

    def transform(self, data_inputs):
        return data_inputs

    def inverse_transform(self, processed_outputs):
        return processed_outputs


class FitCallbackStep(BaseCallbackStep):
    """Call a callback method on fit."""

    def fit(self, data_inputs, expected_outputs=None) -> 'FitCallbackStep':
        """
        Will call the self._callback() with the data being processed and the extra arguments specified.
        Note that here, the data to process is packed into a tuple of (data_inputs, expected_outputs).
        It has no other effect.

        :param data_inputs: the data to process
        :param expected_outputs: the data to process
        :return: self
        """
        self._callback((data_inputs, expected_outputs))
        return self


class TransformCallbackStep(BaseCallbackStep):
    """Call a callback method on transform and inverse transform."""

    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Any]:
        self._callback(data_inputs)

        if self.transform_function is not None:
            return self, self.transform_function(data_inputs)
        return self, data_inputs

    def transform(self, data_inputs):
        """
        Will call the self._callback() with the data being processed and the extra arguments specified.
        It has no other effect.

        :param data_inputs: the data to process
        :return: the same data as input, unchanged (like the Identity class).
        """
        self._callback(data_inputs)
        if self.transform_function is not None:
            return self.transform_function(data_inputs)

        return data_inputs

    def inverse_transform(self, processed_outputs):
        """
        Will call the self._callback() with the data being processed and the extra arguments specified.
        It has no other effect.

        :param processed_outputs: the data to process
        :return: the same data as input, unchanged (like the Identity class).
        """
        self._callback(processed_outputs)
        return processed_outputs


class FitTransformCallbackStep(BaseStep):
    def __init__(self, transform_callback_function=None, fit_callback_function=None, more_arguments: List = tuple(),
                 transform_function=None,
                 hyperparams=None):
        BaseStep.__init__(self, hyperparams)

        if transform_callback_function is None:
            transform_callback_function = TapeCallbackFunction()
        if fit_callback_function is None:
            fit_callback_function = TapeCallbackFunction()

        self.transform_function = transform_function
        self.more_arguments = more_arguments
        self.fit_callback_function: Optional[Callable] = fit_callback_function
        self.transform_callback_function: Optional[Callable] = transform_callback_function

    def fit(self, data_inputs, expected_outputs=None):
        self.fit_callback_function((data_inputs, expected_outputs), *self.more_arguments)
        return self

    def transform(self, data_inputs):
        self.transform_callback_function(data_inputs, *self.more_arguments)
        if self.transform_function is not None:
            return self.transform_function(data_inputs)
        return data_inputs

    def fit_transform(self, data_inputs, expected_outputs=None) -> Tuple['BaseStep', Any]:
        self.fit_callback_function((data_inputs, expected_outputs), *self.more_arguments)
        self.transform_callback_function(data_inputs, *self.more_arguments)
        if self.transform_function is not None:
            return self, self.transform_function(data_inputs)

        return self, data_inputs

    def inverse_transform(self, processed_outputs):
        return processed_outputs

    def clear_callbacks(self):
        cleared_callbacks = {
            self.name: {
                'transform': self.transform_callback_function.data,
                'fit': self.fit_callback_function.data
            }
        }

        self.transform_callback_function.data = []
        self.transform_callback_function.name_tape = []

        self.fit_callback_function.data = []
        self.fit_callback_function.name_tape = []

        return RecursiveDict(cleared_callbacks)


class CallbackWrapper(HandleOnlyMixin, MetaStep):
    """
    A step that calls a callback function for each of his methods : transform, fit, fit_transform, and even inverse_transform.
    To be used with :class:`TapeCallbackFunction`.

    .. code-block:: python

        tape_fit = TapeCallbackFunction()
        tape_transform = TapeCallbackFunction()
        tape_inverse_transform = TapeCallbackFunction()

        callback_wrapper = CallbackWrapper(MultiplyByN(2), tape_transform_preprocessing, tape_fit_preprocessing, tape_inverse_transform_preprocessing)

    .. seealso::
        :class:`~neuraxle.base.HandleOnlyMixin`,
        :class:`~neuraxle.base.MetaStepMixin`,
        :class:`~neuraxle.base.BaseStep`
    """

    def __init__(
            self,
            wrapped,
            transform_callback_function,
            fit_callback_function,
            inverse_transform_callback_function=None,
            more_arguments: List = tuple(),
            hyperparams=None
    ):
        MetaStep.__init__(self, wrapped=wrapped, hyperparams=hyperparams)

        self.inverse_transform_callback_function = inverse_transform_callback_function
        self.more_arguments = more_arguments
        self.fit_callback_function = fit_callback_function
        self.transform_callback_function = transform_callback_function

    def _fit_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseStep', DACT]:
        """
        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: step, data_container
        """
        self.fit_callback_function((data_container.data_inputs, data_container.expected_outputs), *self.more_arguments)
        self.wrapped = self.wrapped.handle_fit(data_container, context)
        return self

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseStep', DACT]:
        """
        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: step, data_container
        """
        self.fit_callback_function((data_container.data_inputs, data_container.expected_outputs), *self.more_arguments)
        self.transform_callback_function(data_container.data_inputs, *self.more_arguments)
        self.wrapped, data_container = self.wrapped.handle_fit_transform(data_container, context)
        return self, data_container

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        """
        :param data_container: data container
        :type data_container: DataContainer
        :param context: execution context
        :type context: ExecutionContext
        :return: step, data_container
        :type: DataContainer
        """
        self.transform_callback_function(data_container.data_inputs, *self.more_arguments)
        return self.wrapped.handle_transform(data_container, context.push(self.wrapped))

    def handle_inverse_transform(self, data_container: DACT, context: CX) -> DACT:
        """
        :param context: execution context
        :type context: ExecutionContext
        :param data_container: data containerj
        :type data_container: DataContainer
        :return: data container
        :rtype: DataContainer
        """
        self.inverse_transform_callback_function(data_container.data_inputs, *self.more_arguments)
        data_container = self.wrapped.handle_inverse_transform(data_container, context.push(self.wrapped))

        return data_container


class TapeCallbackFunction:
    """This class's purpose is to be sent to the callback to accumulate information.

    Example usage:

    .. code-block:: python

        expected_tape = ["1", "2", "3", "a", "b", "4"]
        tape = TapeCallbackFunction()

        p = Pipeline([
            Identity(),
            TransformCallbackStep(tape.callback, ["1"]),
            TransformCallbackStep(tape.callback, ["2"]),
            TransformCallbackStep(tape.callback, ["3"]),
            AddFeatures([
                TransformCallbackStep(tape.callback, ["a"]),
                TransformCallbackStep(tape.callback, ["b"]),
            ]),
            TransformCallbackStep(tape.callback, ["4"]),
            Identity()
        ])
        p.fit_transform(np.ones((1, 1)))

        assert expected_tape == tape.get_name_tape()

    """

    def __init__(self):
        """Initialize the tape (cache lists)."""
        self.data: List = []
        self.name_tape: List[str] = []

    def __call__(self, *args, **kwargs):
        return self.callback(*args, **kwargs)

    def callback(self, data, name: str = ""):
        """
        Will stick the data and name to the tape.

        :param data: data to save
        :param name: name to save (string)
        :return: None
        """
        self.data.append(data)
        self.name_tape.append(name)

    def reset(self):
        """
        Reset callback data.
        :return: None
        """
        self.data = []
        self.name_tape = []

    def get_data(self) -> List:
        """
        Get the data tape

        :return: The list of data.
        """
        return self.data

    def get_name_tape(self) -> List[str]:
        """
        Get the data tape

        :return: The list of names.
        """
        return self.name_tape


class HandleCallbackStep(ForceHandleOnlyMixin, BaseStep):
    def __init__(
            self,
            handle_fit_callback,
            handle_transform_callback,
            handle_fit_transform_callback
    ):
        BaseStep.__init__(self)
        ForceHandleOnlyMixin.__init__(self)

        self.handle_fit_callback = handle_fit_callback
        self.handle_fit_transform_callback = handle_fit_transform_callback
        self.handle_transform_callback = handle_transform_callback

    def _fit_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseStep', DACT]:
        self.handle_fit_callback((data_container, context))
        return self, data_container

    def _transform_data_container(
        self, data_container: DACT, context: CX
    ) -> DACT:
        self.handle_transform_callback((data_container, context))
        return data_container

    def _fit_transform_data_container(
        self, data_container: DACT, context: CX
    ) -> Tuple['BaseStep', DACT]:
        self.handle_fit_transform_callback((data_container, context))
        return self, data_container


class Sleep(BaseTransformer):
    def __init__(self, sleep_time=0.1, hyperparams=None, hyperparams_space=None):
        BaseTransformer.__init__(self, hyperparams=hyperparams, hyperparams_space=hyperparams_space)
        self.sleep_time = sleep_time

    def transform(self, data_inputs):
        time.sleep(self.sleep_time)
        return data_inputs



import pytest
from typing import Callable
from neuraxle.base import MetaStep, NonTransformableMixin
from neuraxle.hyperparams.distributions import Boolean, RandInt
from neuraxle.hyperparams.space import HyperparameterSamples, HyperparameterSpace, RecursiveDict
from neuraxle.pipeline import Pipeline
from neuraxle.steps.loop import StepClonerForEachDataInput
from testing.test_pipeline import SomeStep


SOME_STEP_HP_KEY = 'somestep_hyperparam'
RAND_INT_SOME_STEP = RandInt(-10, 0)
RAND_INT_STEP_CLONER = RandInt(0, 10)

META_STEP_HP = 'metastep_hyperparam'
SOME_STEP_HP = "SomeStep__somestep_hyperparam"
META_STEP_HP_VALUE = 1
SOME_STEP_HP_VALUE = 2

HYPE_SPACE = HyperparameterSpace({
    "a__test": Boolean()
})

HYPE_SAMPLE = HyperparameterSamples({
    "a__test": True
})


class SomeMetaStep(NonTransformableMixin, MetaStep):
    def __init__(self, wrapped):
        MetaStep.__init__(self, wrapped)
        NonTransformableMixin.__init__(self)


class SomeStepInverseTransform(SomeStep):
    def fit_transform(self, data_inputs, expected_outputs=None):
        return self, 'fit_transform'

    def inverse_transform(self, processed_outputs):
        return 'inverse_transform'


def test_step_cloner_should_get_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    hyperparams = p.get_hyperparams()

    assert hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert hyperparams[SOME_STEP_HP] == SOME_STEP_HP_VALUE


def test_step_cloner_should_set_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE


def test_step_cloner_update_hyperparams_should_update_step_cloner_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    p.update_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE + 1,
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE + 1
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE


def test_step_cloner_update_hyperparams_should_update_wrapped_step_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    p.update_hyperparams({
        SOME_STEP_HP: SOME_STEP_HP_VALUE + 1
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE + 1


def test_step_cloner_update_hyperparams_space_should_update_step_cloner_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    update_meta_step_hp_space = RandInt(0, 40)
    p.update_hyperparams_space({
        META_STEP_HP: update_meta_step_hp_space
    })

    assert isinstance(p.hyperparams_space, HyperparameterSpace)
    assert p.hyperparams_space[META_STEP_HP] == update_meta_step_hp_space
    assert p.wrapped.get_hyperparams_space()[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_step_cloner_update_hyperparams_space_should_update_wrapped_step_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    updated_some_step_hp_space = RandInt(0, 400)
    p.update_hyperparams_space({
        SOME_STEP_HP: updated_some_step_hp_space
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert p.wrapped.get_hyperparams_space()[SOME_STEP_HP_KEY] == updated_some_step_hp_space


def test_step_cloner_should_set_steps_hyperparams():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert isinstance(p.get_step().hyperparams, HyperparameterSamples)
    assert p.get_step().get_hyperparams()[SOME_STEP_HP_KEY] == SOME_STEP_HP_VALUE


def test_step_cloner_should_set_steps_hyperparams_space():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_STEP_CLONER,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    assert isinstance(p.get_step().hyperparams_space, HyperparameterSpace)
    assert p.get_step().hyperparams_space[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_step_cloner_should_set_hyperparams_space():
    p = StepClonerForEachDataInput(SomeStep())

    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_STEP_CLONER,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    assert isinstance(p.hyperparams_space, HyperparameterSpace)
    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_STEP_CLONER
    assert p.get_step().hyperparams_space[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_step_cloner_should_get_hyperparams_space():
    p = StepClonerForEachDataInput(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_STEP_CLONER,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    hyperparams_space = p.get_hyperparams_space()

    assert hyperparams_space[META_STEP_HP] == RAND_INT_STEP_CLONER
    assert hyperparams_space[SOME_STEP_HP] == RAND_INT_SOME_STEP


RAND_INT_META_STEP = RandInt(0, 10)


def test_pipeline_should_set_hyperparams():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])

    p.set_hyperparams({
        'hp': 1,
        'step_1__hp': 2,
        'step_2__hp': 3
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams['hp'] == 1
    assert p[0].hyperparams['hp'] == 2
    assert p[1].hyperparams['hp'] == 3


def test_pipeline_should_get_set_config():
    p = Pipeline([
        SomeStep().set_name('step_1').set_config({'c2': 2}),
        SomeStep().set_name('step_2')
    ])

    p.update_config({
        'c1': 1,
        'step_2__c3': 3
    })
    remade_config: RecursiveDict = p.get_config()

    assert isinstance(remade_config, RecursiveDict)
    assert remade_config['c1'] == 1
    assert remade_config['step_1']['c2'] == 2
    assert remade_config['step_2']['c3'] == 3


def test_pipeline_should_not_set_hyperparams_for_unexisting_step():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])

    try:
        p.set_hyperparams({
            'hp': 1,
            'step_3__hp': 2
        })
    except KeyError:
        assert True
    else:
        assert False, "Should raise KeyError on step_3 that doesn't exist."


def test_pipeline_should_get_hyperparams():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])
    p.set_hyperparams({
        'hp': 1,
        'step_1__hp': 2,
        'step_2__hp': 3
    })

    hyperparams = p.get_hyperparams()

    assert isinstance(hyperparams, HyperparameterSamples)
    assert hyperparams['hp'] == 1
    assert hyperparams['step_1__hp'] == 2
    assert hyperparams['step_2__hp'] == 3


def test_pipeline_should_get_hyperparams_space():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])
    p.set_hyperparams_space({
        'hp': RandInt(1, 2),
        'step_1__hp': RandInt(2, 3),
        'step_2__hp': RandInt(3, 4)
    })

    hyperparams_space = p.get_hyperparams_space()

    assert isinstance(hyperparams_space, HyperparameterSpace)

    assert hyperparams_space['hp'].min_included == 1
    assert hyperparams_space['hp'].max_included == 2

    assert hyperparams_space['step_1__hp'].min_included == 2
    assert hyperparams_space['step_1__hp'].max_included == 3

    assert hyperparams_space['step_2__hp'].min_included == 3
    assert hyperparams_space['step_2__hp'].max_included == 4


def test_pipeline_should_update_hyperparams():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])

    p.set_hyperparams({
        'hp': 1,
        'step_1__hp': 2,
        'step_2__hp': 3
    })

    p.update_hyperparams({
        'hp': 4,
        'step_2__hp': 6
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams['hp'] == 4
    assert p[0].hyperparams['hp'] == 2
    assert p[1].hyperparams['hp'] == 6


def test_pipeline_should_update_hyperparams_space():
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])

    p.set_hyperparams_space({
        'hp': RandInt(1, 2),
        'step_1__hp': RandInt(2, 3),
        'step_2__hp': RandInt(3, 4)
    })
    p.update_hyperparams_space({
        'hp': RandInt(4, 6),
        'step_2__hp': RandInt(6, 8)
    })

    assert isinstance(p.hyperparams_space, HyperparameterSpace)

    assert p.hyperparams_space['hp'].min_included == 4
    assert p.hyperparams_space['hp'].max_included == 6

    assert p[0].hyperparams_space['hp'].min_included == 2
    assert p[0].hyperparams_space['hp'].max_included == 3

    assert p[1].hyperparams_space['hp'].min_included == 6
    assert p[1].hyperparams_space['hp'].max_included == 8


@pytest.mark.parametrize('hyperparam_apply_func', [
    'set_hyperparams_space', 'update_hyperparams_space'
])
def test_pipeline_wont_set_hyperparams_space_if_not_distribution_type(hyperparam_apply_func: str):
    p = Pipeline([
        SomeStep().set_name('step_1'),
        SomeStep().set_name('step_2')
    ])
    not_a_distribution = [3, 4]  # will throw error.
    space = {
        'hp': RandInt(1, 2),
        'step_1__hp': RandInt(2, 3),
        'step_2__hp': not_a_distribution
    }
    with pytest.raises(AssertionError):
        hyperparam_apply = getattr(p, hyperparam_apply_func)

        hyperparam_apply(space)


def test_meta_step_mixin_should_get_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams(HyperparameterSamples({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    }))

    hyperparams = p.get_hyperparams()

    assert hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert hyperparams[SOME_STEP_HP] == SOME_STEP_HP_VALUE


def test_meta_step_mixin_should_set_hyperparams():
    p = SomeMetaStep(SomeStep())

    p.set_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    })

    assert isinstance(p.hyperparams, HyperparameterSamples)
    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()['somestep_hyperparam'] == SOME_STEP_HP_VALUE


def test_meta_step_mixin_update_hyperparams_should_update_meta_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams(HyperparameterSamples({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    }))

    p.update_hyperparams({
        META_STEP_HP: META_STEP_HP_VALUE + 1
    })

    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE + 1
    assert p.get_step().get_hyperparams()['somestep_hyperparam'] == SOME_STEP_HP_VALUE


def test_meta_step_mixin_update_hyperparams_should_update_wrapped_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams(HyperparameterSamples({
        META_STEP_HP: META_STEP_HP_VALUE,
        SOME_STEP_HP: SOME_STEP_HP_VALUE
    }))

    p.update_hyperparams({
        SOME_STEP_HP: SOME_STEP_HP_VALUE + 1
    })

    assert p.hyperparams[META_STEP_HP] == META_STEP_HP_VALUE
    assert p.get_step().get_hyperparams()['somestep_hyperparam'] == SOME_STEP_HP_VALUE + 1


def test_meta_step_mixin_update_hyperparams_space_should_update_meta_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    update_meta_step_hp_space = RandInt(0, 100)
    p.update_hyperparams_space({
        META_STEP_HP: update_meta_step_hp_space
    })

    assert p.hyperparams_space[META_STEP_HP] == update_meta_step_hp_space
    assert p.wrapped.get_hyperparams_space()['somestep_hyperparam'] == RAND_INT_SOME_STEP


def test_meta_step_mixin_update_hyperparams_space_should_update_wrapped_step_hyperparams():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    updated_some_step_hp_space = RandInt(0, 100)
    p.update_hyperparams_space({
        SOME_STEP_HP: updated_some_step_hp_space
    })

    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert p.wrapped.get_hyperparams_space()['somestep_hyperparam'] == updated_some_step_hp_space


def test_meta_step_mixin_should_set_hyperparams_space():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    })

    assert p.hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert p.get_step().hyperparams_space[SOME_STEP_HP_KEY] == RAND_INT_SOME_STEP


def test_meta_step_mixin_should_get_hyperparams_space():
    p = SomeMetaStep(SomeStep())
    p.set_hyperparams_space(HyperparameterSpace({
        META_STEP_HP: RAND_INT_META_STEP,
        SOME_STEP_HP: RAND_INT_SOME_STEP
    }))

    hyperparams_space = p.get_hyperparams_space()

    assert hyperparams_space[META_STEP_HP] == RAND_INT_META_STEP
    assert hyperparams_space[SOME_STEP_HP] == RAND_INT_SOME_STEP


def test_get_set_params_base_step():
    s = SomeStep()

    s.set_params(learning_rate=0.1)
    hyperparams = s.get_params()

    assert hyperparams == {"learning_rate": 0.1}



"""
Neuraxle's SQLAlchemy Hyperparameter Repository Classes
=================================================
Data objects and related repositories used by AutoML, SQL version.

Classes are splitted like this for the AutoML:
- Projects
- Clients
- Rounds (runs)
- Trials
- TrialSplits
- MetricResults

..
    Copyright 2021, Neuraxio Inc.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.


"""
from typing import List

from neuraxle.base import TrialStatus
from neuraxle.metaopt.data.aggregates import Round, Trial
from neuraxle.metaopt.data.vanilla import HyperparamsRepository
from sqlalchemy import (Boolean, DateTime, MetaData, PickleType, String, Table,
                        create_engine)
from sqlalchemy.testing.schema import Column


def get_database_path(user: str, password: str, host: str, dialect: str, driver: str = ''):
    """

    :param user:
    :param password:
    :param dialect: e.g. mysql
    :param driver: (Optional) e.g. "pymysql"
    :return:
    """
    return "dialect[+driver]://user:password@host/dbname"


def get_database_path_sqlite(path):
    return f"sqlite:///{path}"


meta = MetaData()

trial_register = Table(
    'trials', meta,
    Column('id', String, primary_key=True),
    Column('status', String),
    Column('hyperparams', PickleType),
    Column('main_metric_name', String),
    Column('start_time', DateTime),
    Column('end_time', DateTime),
    Column('error', String),
    Column('error_traceback', String)
)

metrics_register = Table(
    'metrics', meta,
    Column('name', String),
    Column('is_higher_score_better', Boolean)
)

metrics_measurements = Table(
    'measurements', meta,
    Column('trial_id', String, foreign_key=True),
    Column('is_validation', Boolean),
    Column('value', PickleType)  # Should contain a list of measurements
)


class InDatabaseHyperparamRepository(HyperparamsRepository):
    def __init__(self, db_path):
        self.engine = create_engine(db_path, echo=True)

    def create_db(self):
        meta.create_all(self.engine)

    def _execute(self, *expressions: List[str]):
        """
        :param expressions:
        :return:
        :rtype Union[ResultProxy, List[ResultProxy]]
        """
        if len(expressions) == 0:
            raise ValueError("No expressions provided for execution!")

        result = []
        with self.engine.connect() as conn:
            for expr in expressions:
                result.append(conn.execute(expr))

        return result[0] if len(result) == 1 else result

    def load_trials(self, status: 'TrialStatus') -> Round:
        """
        Load all hyperparameter trials with their corresponding score.
        Sorted by creation date.

        :return: Trials (hyperparams, scores)
        """
        select = trial_register.select()
        if status is not None:
            select = select.where(status=TrialStatus.value)
        res = self._execute(select)

        # trials = Round()

        for _ in res:  # loop on rows
            raise NotImplementedError()
            # trial = Trial()
            # trials.append(trial)

    def _save_trial(self, trial: 'Trial'):
        """
        save trial.

        :param trial: trial to save.
        :return:
        """
        raise NotImplementedError()

    def _insert_new_trial(self, trial: Trial):
        trial_hash = self.get_trial_id(trial.hyperparams)
        return trial_register.insert().value(id=trial_hash, status=TrialStatus.PLANNED)

    def new_trial(self, auto_ml_container):
        """
        Create a new trial with the best next hyperparams.

        :param context:
        :param auto_ml_container: auto ml data container
        :return: trial
        """
        trial = HyperparamsRepository.new_trial(self, auto_ml_container)
        query = self._insert_new_trial(trial)
        self._execute(query)
        return trial



import numpy as np

from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.steps.flow import OptionalStep
from neuraxle.steps.numpy import MultiplyByN


def test_optional_should_disable_wrapped_step_when_disabled():
    p = OptionalStep(MultiplyByN(2), nullified_return_value=[]).set_hyperparams(HyperparameterSamples({
        'enabled': False
    }))
    data_inputs = np.array(list(range(10)))

    outputs = p.transform(data_inputs)

    assert outputs == []


def test_optional_should_enable_wrapped_step_when_enabled():
    p = OptionalStep(MultiplyByN(2), nullified_return_value=[]).set_hyperparams(HyperparameterSamples({
        'enabled': True
    }))
    data_inputs = np.array(list(range(10)))

    outputs = p.transform(data_inputs)

    assert np.array_equal(outputs, data_inputs * 2)




import numpy as np
import pytest

from neuraxle.base import CX, BaseTransformer, ForceHandleMixin
from neuraxle.data_container import DataContainer as DACT
from neuraxle.hyperparams.space import HyperparameterSamples
from neuraxle.pipeline import Pipeline
from neuraxle.steps.misc import FitCallbackStep, TapeCallbackFunction
from neuraxle.steps.numpy import MultiplyByN
from neuraxle.steps.output_handlers import OutputTransformerWrapper, InputAndOutputTransformerWrapper, \
    InputAndOutputTransformerMixin


class MultiplyByNInputAndOutput(BaseTransformer):
    def __init__(self, multiply_by=1):
        super().__init__(hyperparams=HyperparameterSamples({'multiply_by': multiply_by}))

    def transform(self, data_inputs):
        data_inputs, expected_outputs = data_inputs

        if not isinstance(data_inputs, np.ndarray):
            data_inputs = np.array(data_inputs)

        if not isinstance(expected_outputs, np.ndarray):
            expected_outputs = np.array(expected_outputs)

        return data_inputs * self.hyperparams['multiply_by'], expected_outputs * self.hyperparams['multiply_by']

    def inverse_transform(self, processed_outputs):
        processed_outputs, expected_outputs = processed_outputs

        if not isinstance(processed_outputs, np.ndarray):
            processed_outputs = np.array(processed_outputs)

        if not isinstance(expected_outputs, np.ndarray):
            expected_outputs = np.array(expected_outputs)

        return processed_outputs / self.hyperparams['multiply_by'], expected_outputs / self.hyperparams['multiply_by']


def test_output_transformer_wrapper_should_fit_with_data_inputs_and_expected_outputs_as_data_inputs():
    tape = TapeCallbackFunction()
    p = OutputTransformerWrapper(FitCallbackStep(tape))
    data_inputs, expected_outputs = _create_data_source((10, 10))

    p.fit(data_inputs, expected_outputs)

    assert np.array_equal(tape.data[0][0], expected_outputs)
    for i in range(10):
        assert tape.data[0][1][i] is None


def test_output_transformer_wrapper_should_fit_transform_with_data_inputs_and_expected_outputs():
    tape = TapeCallbackFunction()
    p = OutputTransformerWrapper(Pipeline([MultiplyByN(2), FitCallbackStep(tape)]))
    data_inputs, expected_outputs = _create_data_source((10, 10))

    p, data_container = p.handle_fit_transform(DACT(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    ), CX())

    assert np.array_equal(data_container.data_inputs, data_inputs)
    assert np.array_equal(data_container.expected_outputs, expected_outputs * 2)
    assert np.array_equal(tape.data[0][0], expected_outputs * 2)
    for i in range(10):
        assert tape.data[0][1][i] is None


def test_output_transformer_wrapper_should_transform_with_data_inputs_and_expected_outputs():
    p = OutputTransformerWrapper(MultiplyByN(2))
    data_inputs, expected_outputs = _create_data_source((10, 10))

    data_container = p.handle_transform(DACT(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    ), CX())

    assert np.array_equal(data_container.data_inputs, data_inputs)
    assert np.array_equal(data_container.expected_outputs, expected_outputs * 2)


def test_input_and_output_transformer_wrapper_should_fit_with_data_inputs_and_expected_outputs_as_data_inputs():
    tape = TapeCallbackFunction()
    p = InputAndOutputTransformerWrapper(FitCallbackStep(tape))
    data_inputs, expected_outputs = _create_data_source((10, 10))

    p.fit(data_inputs, expected_outputs)

    assert np.array_equal(tape.data[0][0][0], data_inputs)
    assert np.array_equal(tape.data[0][0][1], expected_outputs)


def test_input_and_output_transformer_wrapper_should_fit_transform_with_data_inputs_and_expected_outputs():
    tape = TapeCallbackFunction()
    p = InputAndOutputTransformerWrapper(Pipeline([MultiplyByNInputAndOutput(2), FitCallbackStep(tape)]))
    data_inputs, expected_outputs = _create_data_source((10, 10))

    p, data_container = p.handle_fit_transform(DACT(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    ), CX())

    assert np.array_equal(data_container.data_inputs, data_inputs * 2)
    assert np.array_equal(data_container.expected_outputs, expected_outputs * 2)
    assert np.array_equal(tape.data[0][0][0], data_inputs * 2)
    assert np.array_equal(tape.data[0][0][1], expected_outputs * 2)


def test_input_and_output_transformer_wrapper_should_transform_with_data_inputs_and_expected_outputs():
    p = InputAndOutputTransformerWrapper(MultiplyByNInputAndOutput(2))
    data_inputs, expected_outputs = _create_data_source((10, 10))

    data_container = p.handle_transform(DACT(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    ), CX())

    assert np.array_equal(data_container.data_inputs, data_inputs * 2)
    assert np.array_equal(data_container.expected_outputs, expected_outputs * 2)


class ChangeLenDataInputs(BaseTransformer):
    """
    This should raise an error because it does not return the same length for data_inputs and expected_outputs
    """

    def __init__(self):
        super().__init__()

    def transform(self, data_inputs):
        data_inputs, expected_outputs = data_inputs
        return data_inputs[0:int(len(data_inputs) / 2)], expected_outputs


class ChangeLenDataInputsAndExpectedOutputs(BaseTransformer):
    """
    This should raise an error because ids are not changed to fit the new length of data_inputs and expected_outputs
    """

    def __init__(self):
        BaseTransformer.__init__(self)

    def transform(self, data_inputs):
        data_inputs, expected_outputs = data_inputs
        return data_inputs[0:int(len(data_inputs) / 2)], expected_outputs[0:int(len(data_inputs) / 2)]


class DoubleData(ForceHandleMixin, BaseTransformer):
    """
    This should double the data given in entry. Expects to be wrapped in an InputAndOutputTransformerWrapper.
    """

    def __init__(self):
        BaseTransformer.__init__(self)
        ForceHandleMixin.__init__(self)

    def _transform_data_container(self, data_container: DACT, context: CX) -> DACT:
        di, eo = data_container.data_inputs
        return DACT(data_inputs=(di[0].tolist()*2, eo[0].tolist()*2),
                             ids=data_container.ids*2)


def test_input_and_output_transformer_wrapper_should_not_return_a_different_amount_of_data_inputs_and_expected_outputs():
    with pytest.raises(AssertionError):
        p = InputAndOutputTransformerWrapper(ChangeLenDataInputs())
        data_inputs, expected_outputs = _create_data_source((10, 10))

        p.handle_transform(DACT(
            data_inputs=data_inputs,
            expected_outputs=expected_outputs
        ), CX())


def test_input_and_output_transformer_wrapper_should_raise_an_assertion_error_if_ids_have_not_been_resampled_correctly():
    with pytest.raises(AssertionError) as e:
        p = InputAndOutputTransformerWrapper(ChangeLenDataInputsAndExpectedOutputs())
        data_inputs, expected_outputs = _create_data_source((10, 10))

        p.handle_transform(DACT(
            data_inputs=data_inputs,
            expected_outputs=expected_outputs
        ), CX())


def test_data_doubler():
    p = InputAndOutputTransformerWrapper(DoubleData())
    data_inputs, expected_outputs = _create_data_source((10, 10))

    out = p.handle_transform(DACT(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    ), CX())

    doubled_length = len(out.data_inputs)
    assert doubled_length == 2*len(data_inputs)
    assert doubled_length == len(out.expected_outputs)
    assert doubled_length == len(out.ids)


def _create_data_source(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    expected_outputs = np.random.random(shape).astype(np.float32)
    return data_inputs, expected_outputs



import os
import shutil
from abc import ABC, abstractmethod

import numpy as np
import pytest
from neuraxle.base import (BaseService, BaseStep, BaseTransformerT,
                           CX, ForceHandleIdentity, Identity,
                           NonFittableMixin, StepWithContext)
from neuraxle.data_container import DataContainer as DACT
from neuraxle.metaopt.auto_ml import AutoML, RandomSearchSampler
from neuraxle.metaopt.callbacks import ScoringCallback
from neuraxle.metaopt.data.json_repo import HyperparamsOnDiskRepository
from neuraxle.metaopt.validation import GridExplorationSampler, ValidationSplitter
from neuraxle.pipeline import Pipeline
from sklearn.metrics import mean_squared_error


class SomeBaseService(BaseService):
    @abstractmethod
    def service_method(self, data):
        pass


class SomeService(SomeBaseService):
    def service_method(self, data):
        self.data = data


class AnotherService(BaseService):
    def __init__(self):
        super().__init__()

    def _setup(self, context: 'CX' = None) -> BaseTransformerT:
        return BaseService._setup(self, context=context)


class SomeStep(ForceHandleIdentity):
    def _will_process(self, data_container: DACT, context: CX):
        data_container, context = super()._will_process(data_container, context)
        service = context.get_service(SomeBaseService)
        service.service_method(data_container.data_inputs)
        return data_container, context


class RegisterServiceDynamically(ForceHandleIdentity):
    def _will_process(self, data_container: DACT, context: CX):
        context.register_service(SomeBaseService, SomeService())
        return data_container, context


class SomeStepWithFailedAssertion(NonFittableMixin, BaseStep):
    def __init__(self):
        BaseStep.__init__(self)
        NonFittableMixin.__init__(self)

    # Create a step that will self._assert false and raise an exception and ensure it raised.
    def transform(self, data_inputs):
        for di in data_inputs:
            self._assert(di, f"{di}")
        return None


def test_context_assertions_raises_when_it_throws(tmpdir):
    assertion_was_raised = False

    try:
        SomeStepWithFailedAssertion().transform([True, True, False])
    except AssertionError:
        assertion_was_raised = True

    assert assertion_was_raised


def test_context_assertions_passes_when_it_passes(tmpdir):
    assertion_was_raised = False

    try:
        SomeStepWithFailedAssertion().transform([True, True, True])
    except AssertionError:
        assertion_was_raised = True

    assert not assertion_was_raised


def test_with_context_should_inject_dependencies_properly(tmpdir):
    data_inputs = np.array([0, 1, 2, 3])
    context = CX(root=tmpdir)
    service = SomeService()
    context.set_service_locator({SomeBaseService: service})
    p = Pipeline([
        SomeStep().assert_has_services(SomeBaseService),
        SomeStep().assert_has_services_at_execution(SomeBaseService)
    ]).with_context(context=context)

    p.transform(data_inputs=data_inputs)

    assert np.array_equal(service.data, data_inputs)


def test_with_context_should_fail_at_init_when_services_are_missing(tmpdir):
    context = CX(root=tmpdir)
    p = Pipeline([
        SomeStep().assert_has_services(SomeBaseService)
    ]).with_context(context=context)

    data_inputs = np.array([0, 1, 2, 3])
    with pytest.raises(AssertionError) as exception_info:
        p.transform(data_inputs=data_inputs)

    assert 'Expected context to have service of type SomeBaseService' in exception_info.value.args[0]


def test_localassert_should_assert_dependencies_properly_at_exec(tmpdir):
    data_inputs = np.array([0, 1, 2, 3])
    context = CX(root=tmpdir)
    RegisterServiceDynamically().handle_transform(DACT(data_inputs), context)
    p = Pipeline([Pipeline([
        SomeStep().assert_has_services_at_execution(SomeBaseService)
    ])]).with_context(context=context)

    p.transform(data_inputs)
    service = p.context.get_service(SomeBaseService)
    assert np.array_equal(service.data, data_inputs)


def test_localassert_should_fail_when_services_are_missing_at_exec(tmpdir):
    context = CX(root=tmpdir)
    p = Pipeline([
        SomeStep().assert_has_services_at_execution(SomeBaseService)
    ]).with_context(context=context)
    data_inputs = np.array([0, 1, 2, 3])

    with pytest.raises(AssertionError) as exception_info:
        p.transform(data_inputs=data_inputs)

    assert 'Expected context to have service of type SomeBaseService' in exception_info.value.args[0]


def _make_autoML_loop(tmpdir, p: Pipeline):
    return AutoML(
        pipeline=p,
        hyperparams_optimizer=GridExplorationSampler(1),
        validation_splitter=ValidationSplitter(0.20),
        scoring_callback=ScoringCallback(mean_squared_error, higher_score_is_better=False),
        n_trials=1,
        refit_best_trial=True,
        epochs=1,
        continue_loop_on_error=False
    )


class TestServiceAssertion:

    def _setup(self, tmpdir):
        self.tmpdir = str(tmpdir)
        self.tmpdir_hp = self.tmpdir + "_hp"

    def teardown(self):
        shutil.rmtree(self.tmpdir, ignore_errors=True)
        shutil.rmtree(self.tmpdir + "_hp", ignore_errors=True)

    def test_step_with_context_saver_only_saves_wrapped(self, tmpdir):
        self._setup(tmpdir)
        pipeline_name = 'testname'
        context = CX(tmpdir).set_service_locator({SomeBaseService: SomeService()})
        p = Pipeline([
            SomeStep().assert_has_services(SomeBaseService)
        ]).set_name(pipeline_name).with_context(context=context)

        p.save(context, full_dump=True)

        p: Pipeline = CX(tmpdir).load(os.path.join(pipeline_name))
        assert isinstance(p, Pipeline)

    def test_auto_ml_should_inject_dependencies_properly(self, tmpdir):
        self._setup(tmpdir)
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        p = Pipeline([
            SomeStep().assert_has_services(SomeBaseService),
            SomeStep().assert_has_services_at_execution(SomeBaseService)
        ])
        context = CX(root=self.tmpdir)
        service = SomeService()
        context.set_service_locator({SomeBaseService: service})

        auto_ml: AutoML = _make_autoML_loop(self.tmpdir_hp, p)
        auto_ml: StepWithContext = auto_ml.with_context(context=context)
        assert isinstance(auto_ml, StepWithContext)
        auto_ml.fit(data_inputs, expected_outputs)

        assert np.array_equal(service.data, data_inputs)

    def test_auto_ml_should_fail_at_init_when_services_are_missing(self, tmpdir):
        self._setup(tmpdir)
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        p = Pipeline([
            RegisterServiceDynamically(),
            SomeStep().assert_has_services(SomeBaseService),
        ])

        context = CX(root=self.tmpdir)

        auto_ml: AutoML = _make_autoML_loop(self.tmpdir_hp, p)
        auto_ml: StepWithContext = auto_ml.with_context(context=context)
        assert isinstance(auto_ml, StepWithContext)

        with pytest.raises(AssertionError) as exception_info:
            auto_ml.fit(data_inputs, expected_outputs)

        assert 'Expected context to have service of type SomeBaseService but it did not.' in exception_info.value.args[
            0]

    def test_auto_ml_should_fail_at_exec_when_services_are_missing(self, tmpdir):
        self._setup(tmpdir)
        data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        expected_outputs = data_inputs * 2
        p = Pipeline([
            SomeStep().assert_has_services_at_execution(SomeBaseService),
        ])
        context = CX(root=self.tmpdir)

        auto_ml: AutoML = _make_autoML_loop(self.tmpdir_hp, p)
        auto_ml: StepWithContext = auto_ml.with_context(context=context)
        assert isinstance(auto_ml, StepWithContext)

        with pytest.raises(AssertionError) as exception_info:
            auto_ml.fit(data_inputs, expected_outputs)
        assert 'Expected context to have service of type SomeBaseService but it did not.' in exception_info.value.args[
            0]


def test_context_can_access_services_from_apply_method(tmpdir):
    data_inputs = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    expected_outputs = data_inputs * 2

    cx = CX(tmpdir, services={
        SomeBaseService: SomeService(),
        AnotherService: AnotherService()
    })
    p = Pipeline([
        Identity().assert_has_services(AnotherService),
        SomeStep().assert_has_services_at_execution(SomeBaseService),
    ]).with_context(context=cx)

    for i in cx.services.values():
        assert isinstance(i, BaseService)
        assert i.get_config().to_flat_dict() == dict()

    p.fit(data_inputs, expected_outputs)

    service = cx.get_service(SomeBaseService)
    assert np.array_equal(service.data, data_inputs)

    for i in cx.services.values():
        assert i.get_config().to_flat_dict() == dict()



import numpy as np
import pytest

from neuraxle.steps.column_transformer import ColumnSelector2D, ColumnsSelectorND, NumpyColumnSelector2D


@pytest.mark.parametrize('column_selector_2d_class', [ColumnSelector2D, NumpyColumnSelector2D])
def test_column_selector_2d_should_select_range(column_selector_2d_class):
    step = column_selector_2d_class(range(0, 10))
    data_inputs, expected_outputs = _create_data_source((20, 20))

    outputs = step.transform(data_inputs)

    assert np.array_equal(outputs, data_inputs[..., :10])


@pytest.mark.parametrize('column_selector_2d_class', [ColumnSelector2D, NumpyColumnSelector2D])
def test_column_selector_2d_should_select_int(column_selector_2d_class):
    step = column_selector_2d_class(10)
    data_inputs, expected_outputs = _create_data_source((20, 20))

    outputs = step.transform(data_inputs)

    expected_data_inputs = np.expand_dims(data_inputs[..., 10], axis=-1)
    assert np.array_equal(outputs, expected_data_inputs)


@pytest.mark.parametrize('column_selector_2d_class', [ColumnSelector2D, NumpyColumnSelector2D])
def test_column_selector_2d_should_select_slice(column_selector_2d_class):
    step = column_selector_2d_class(slice(0, 10, 1))
    data_inputs, expected_outputs = _create_data_source((20, 20))

    outputs = step.transform(data_inputs)

    assert np.array_equal(outputs, data_inputs[..., :10])


@pytest.mark.parametrize('column_selector_2d_class', [ColumnSelector2D, NumpyColumnSelector2D])
def test_column_selector_2d_should_select_list_of_indexes(column_selector_2d_class):
    step = column_selector_2d_class([0, 1, 2])
    data_inputs, expected_outputs = _create_data_source((20, 20))

    outputs = step.transform(data_inputs)

    assert np.array_equal(outputs, data_inputs[..., :3])


def test_column_selector_nd_should_transform_with_column_selector_2d():
    step = ColumnsSelectorND(0, n_dimension=2)
    data_inputs, expected_outputs = _create_data_source((20, 20))

    outputs = step.transform(data_inputs)

    assert np.array_equal(outputs, np.expand_dims(data_inputs[..., 0], axis=-1))


def _create_data_source(shape):
    data_inputs = np.random.random(shape).astype(np.float32)
    expected_outputs = np.random.random(shape).astype(np.float32)
    return data_inputs, expected_outputs



import pytest

from neuraxle.data_container import DACT, AbsentValuesNullObject
import numpy as np


class LoadableItem:
    def __init__(self):
        self.loaded = False

    def load(self) -> 'LoadableItem':
        self.loaded = True
        return self

    def is_loaded(self):
        return self.loaded


class SomeLazyLoadableCollection:
    def __init__(self, inner_list):
        self.inner_list = inner_list
        self.iterations = 0

    def __iter__(self):
        for item in self.inner_list:
            yield item.load()

    def __getitem__(self, item):
        return SomeLazyLoadableCollection([
            item.load()
            for item in self.inner_list[item]
        ])

    def __len__(self):
        return len(self.inner_list)


def test_data_container_minibatch_should_be_lazy_and_use_getitem_when_data_is_lazy_loadable():
    items = [LoadableItem() for _ in range(10)]
    data_inputs = SomeLazyLoadableCollection(items)
    expected_outputs = SomeLazyLoadableCollection([LoadableItem() for _ in range(10)])
    data_container = DACT(
        data_inputs=data_inputs,
        expected_outputs=expected_outputs
    )

    i = 0
    batch_size = 2
    for batch in data_container.minibatches(batch_size=batch_size):
        assert len(batch) == batch_size
        assert all(item.is_loaded() for item in data_inputs.inner_list[:(i * batch_size)])
        for y in range((i + 1) * batch_size, len(data_inputs)):
            assert not items[y].is_loaded()
        i += 1


@pytest.mark.parametrize('batch_size,include_incomplete_pass,default_value,expected_data_containers', [
    (3, False, None, [
        DACT(ids=[0, 1, 2], data_inputs=[0, 1, 2], expected_outputs=[10, 11, 12]),
        DACT(ids=[3, 4, 5], data_inputs=[3, 4, 5], expected_outputs=[13, 14, 15]),
        DACT(ids=[6, 7, 8], data_inputs=[6, 7, 8], expected_outputs=[16, 17, 18]),
    ]),
    (3, True, 0, [
        DACT(ids=[0, 1, 2], data_inputs=[0, 1, 2], expected_outputs=[10, 11, 12]),
        DACT(ids=[3, 4, 5], data_inputs=[3, 4, 5], expected_outputs=[13, 14, 15]),
        DACT(ids=[6, 7, 8], data_inputs=[6, 7, 8], expected_outputs=[16, 17, 18]),
        DACT(ids=[0, 1, 2], data_inputs=[9, 0, 0], expected_outputs=[19, 0, 0])
    ]),
    (3, True, AbsentValuesNullObject(), [
        DACT(ids=[0, 1, 2], data_inputs=[0, 1, 2], expected_outputs=[10, 11, 12]),
        DACT(ids=[3, 4, 5], data_inputs=[3, 4, 5], expected_outputs=[13, 14, 15]),
        DACT(ids=[6, 7, 8], data_inputs=[6, 7, 8], expected_outputs=[16, 17, 18]),
        DACT(ids=[9], data_inputs=[9], expected_outputs=[19])
    ])
])
def test_data_container_batching(batch_size, include_incomplete_pass, default_value, expected_data_containers):
    data_container = DACT(
        ids=[str(i) for i in range(10)],
        data_inputs=np.array(list(range(10))),
        expected_outputs=np.array(list(range(10, 20)))
    )

    # When
    data_containers = []
    for dc in data_container.minibatches(
        batch_size=batch_size,
        keep_incomplete_batch=include_incomplete_pass,
        default_value_data_inputs=default_value
    ):
        data_containers.append(dc)

    # Then
    assert len(expected_data_containers) == len(data_containers)
    for expected_data_container, actual_data_container in zip(expected_data_containers, data_containers):
        np.array_equal(expected_data_container.ids, actual_data_container.ids)
        np.array_equal(expected_data_container.data_inputs, actual_data_container.data_inputs)
        np.array_equal(expected_data_container.expected_outputs, actual_data_container.expected_outputs)



import os

import numpy as np
from joblib import dump
from py._path.local import LocalPath
from pprint import pprint

from neuraxle.hyperparams.space import RecursiveDict
from neuraxle.base import CX, StepWithContext, TruncableJoblibStepSaver
from neuraxle.pipeline import Pipeline
from neuraxle.steps.numpy import MultiplyByN

OUTPUT = "OUTPUT"
ROOT = 'Pipeline'
PIPELINE_2 = 'Pipeline2'
SOME_STEPS = ['some_step0', 'some_step1', 'some_step2']

EXPECTED_OUTPUTS = [0, 48, 96, 144, 192, 240, 288, 336, 384, 432]


def create_some_step_path(tmpdir, step_no=0, create_dir=False):
    if step_no == 0:
        path1 = os.path.join(tmpdir, ROOT, SOME_STEPS[step_no])
    else:
        path1 = os.path.join(tmpdir, ROOT, PIPELINE_2, SOME_STEPS[step_no])
    if create_dir:
        os.makedirs(path1)
    path2 = os.path.join(path1, '{0}.joblib'.format(SOME_STEPS[step_no]))
    return path2


def create_pipeline2_path(tmpdir, create_dir=False):
    path1 = os.path.join(tmpdir, ROOT, PIPELINE_2)
    if create_dir:
        os.makedirs(path1)
    path2 = os.path.join(path1, '{0}.joblib'.format(PIPELINE_2))
    return path2


def create_root_path(tmpdir, create_dir=False):
    path1 = os.path.join(tmpdir, ROOT)
    if create_dir and not os.path.exists(os.path.join(tmpdir, ROOT)):
        os.makedirs(path1)
    path2 = os.path.join(path1, '{0}.joblib'.format(ROOT))
    return path2


def test_nested_pipeline_fit_transform_should_save_some_fitted_pipeline_steps(tmpdir: LocalPath):
    p: StepWithContext = create_pipeline(tmpdir)

    p, outputs = p.fit_transform(np.array(range(10)), np.array(range(10)))
    p.save()

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)
    saved_paths = [
        create_root_path(tmpdir), create_pipeline2_path(tmpdir),
        create_some_step_path(tmpdir, step_no=0), create_some_step_path(tmpdir, step_no=1),
        create_some_step_path(tmpdir, step_no=2)
    ]
    for path in saved_paths:
        assert os.path.exists(path), path


def test_pipeline_transform_should_not_save_steps(tmpdir: LocalPath):
    p: StepWithContext = create_pipeline(tmpdir)

    outputs = p.transform(np.array(range(10)))
    p.wrapped.save(CX(tmpdir), full_dump=False)

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)
    not_saved_paths = [
        create_root_path(tmpdir), create_pipeline2_path(tmpdir), create_some_step_path(tmpdir, step_no=0),
        create_some_step_path(tmpdir, step_no=1), create_some_step_path(tmpdir, step_no=2)]
    for path in not_saved_paths:
        assert not os.path.exists(path), path


def test_pipeline_fit_should_save_all_fitted_pipeline_steps(tmpdir: LocalPath):
    p: StepWithContext = create_pipeline(tmpdir)

    p = p.fit(np.array(range(10)), np.array(range(10)))
    p.save()

    saved_paths = [
        create_root_path(tmpdir), create_pipeline2_path(tmpdir),
        create_some_step_path(tmpdir, step_no=0), create_some_step_path(tmpdir, step_no=1),
        create_some_step_path(tmpdir, step_no=2)
    ]
    for path in saved_paths:
        assert os.path.exists(path), path


def test_pipeline_fit_transform_should_load_all_pipeline_steps(tmpdir: LocalPath):
    p = given_saved_pipeline(tmpdir)

    p, outputs = p.fit_transform(np.array(range(10)), np.array(range(10)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_pipeline_transform_should_load_all_pipeline_steps(tmpdir: LocalPath):
    p = given_saved_pipeline(tmpdir)

    outputs = p.transform(np.array(range(10)))

    assert np.array_equal(outputs, EXPECTED_OUTPUTS)


def test_pipeline_fit_should_load_all_pipeline_steps(tmpdir: LocalPath):
    p = given_saved_pipeline(tmpdir)

    p = p.fit(np.array(range(10)), np.array(range(10)))

    assert p.wrapped[SOME_STEPS[0]].hyperparams['multiply_by'] == 2
    assert p.wrapped[PIPELINE_2][SOME_STEPS[1]].hyperparams['multiply_by'] == 4
    assert p.wrapped[PIPELINE_2][SOME_STEPS[2]].hyperparams['multiply_by'] == 6


def given_saved_pipeline(tmpdir: LocalPath) -> Pipeline:
    path = create_root_path(tmpdir, True)
    p = Pipeline([]).set_name(ROOT).with_context(CX(tmpdir)).with_context(CX(tmpdir))
    dump(p, path)

    pipeline_2 = Pipeline([]).set_name(PIPELINE_2).with_context(CX(tmpdir))
    pipeline_2.sub_steps_savers = [
        (SOME_STEPS[0], []),
        (SOME_STEPS[1], []),
    ]
    dump(pipeline_2, create_pipeline2_path(tmpdir, True))

    given_saved_some_step(multiply_by=2, step_no=0, path=create_some_step_path(tmpdir, step_no=0, create_dir=True))
    given_saved_some_step(multiply_by=4, step_no=1, path=create_some_step_path(tmpdir, step_no=1, create_dir=True))
    given_saved_some_step(multiply_by=6, step_no=2, path=create_some_step_path(tmpdir, step_no=2, create_dir=True))

    p = create_pipeline(tmpdir)

    return p


def create_pipeline(tmpdir) -> StepWithContext:
    return Pipeline([
        (SOME_STEPS[0], MultiplyByN(multiply_by=2)),
        (PIPELINE_2, Pipeline([
            (SOME_STEPS[1], MultiplyByN(multiply_by=4)),
            (SOME_STEPS[2], MultiplyByN(multiply_by=6))
        ]))
    ]).set_name(ROOT).with_context(CX(tmpdir))


def given_saved_some_step(multiply_by, step_no, path):
    some_step1 = MultiplyByN(multiply_by=multiply_by)
    some_step1.name = SOME_STEPS[step_no]
    dump(some_step1, path)
